# %%
# This document serves multiple purposes:
#   - A reproduction of the experiments in the Paren Balancer causal scrubbing post
#   - A causal scrubbing demo
#   - A set of exercises for remix (hense the `if "SOLUTION"` blocks, and instructions)

# %% [markdown]
"""
# Day 4b: Paren Balancer Causal Scrubbing

To start, please read this [less wrong post](https://www.lesswrong.com/s/h95ayYYwMebGEYN5y/p/kjudfaQazMmC74SbF).

We will replicate the experiments today.

<!-- toc -->
"""
import uuid
from typing import Optional, Tuple

import numpy as np
import torch
from torch.nn.functional import binary_cross_entropy_with_logits

import rust_circuit as rc
import rust_circuit.optional as op
from rust_circuit.algebric_rewrite import residual_rewrite, split_to_concat
from rust_circuit.causal_scrubbing.experiment import (
    Experiment,
    ExperimentCheck,
    ExperimentEvalSettings,
    ScrubbedExperiment,
)
from rust_circuit.causal_scrubbing.hypothesis import (
    Correspondence,
    ExactSampler,
    FuncSampler,
    InterpNode,
    UncondSampler,
    chain_excluding,
    corr_root_matcher,
)
from rust_circuit.demos.notebook_testing import NotebookInTesting
from rust_circuit.model_rewrites import To, configure_transformer
from rust_circuit.module_library import load_model_id
from rust_circuit.py_utils import I
from rust_circuit.ui.ui import circuit_graph_ui, ui_default_hidden_matcher

from . import test_paren_balancer_exercises as tests
from .setup import ParenDataset, ParenTokenizer, get_h00_open_vector

ui_hidden_matcher = ui_default_hidden_matcher
# %%
# define model
MAIN = __name__ == "__main__"

SEQ_LEN = 42
NUM_EXAMPLES = 4000
MODEL_ID = "jun9_paren_balancer"

PRINT_CIRCUITS = True
ACTUALLY_RUN = True
SLOW_EXPERIMENTS = True
DEFAULT_CHECKS: ExperimentCheck = True
EVAL_DEVICE = "cuda:0"
# If you have less memory, you will want to reduce this and also add a batch_size
MAX_MEMORY = 30_000_000_000
BATCH_SIZE = None

if NotebookInTesting.currently_in_notebook_test:
    EVAL_DEVICE = "cpu"
    SLOW_EXPERIMENTS = False
    MAIN = True  # a bit of a lie but we want to actually run things

# %% [markdown]
"""
## Setup 
No exercises here! It may be helpful to read over the code, however.

### Circuit loading
If any of these operations confuse you, try printing out the circuit before and after!

Step 1: Initial loading
"""

circ_dict, _, model_info = load_model_id(MODEL_ID)
circuit = circ_dict["t.bind_w"]

#%% [markdown]
"""
Step 2: We bind the model to an input by attaching a placeholder symbol input named "tokens" to the model. We then specify the attention mask that prevents attending to padding depends on this tokens array.

We use one hot tokens as this makes defining the attention mask a simple indexing operation.

The symbol has a random fixed uuid (fixed as this gives consistency when comparing in tests).
"""
toks_uuid = uuid.UUID("ce34280e-169f-40bd-b78e-8adeb4274aba")
tokens_arr = rc.Symbol((SEQ_LEN, ParenTokenizer.vocab_size), uuid=toks_uuid, name="tokens")
tok_embeds = rc.Einsum.from_fancy_string(
    "seqlen vocab_size, vocab_size hidden -> seqlen hidden", tokens_arr, circ_dict["t.w.tok_embeds"], name="tok_embeds"
)
attn_mask = rc.Add.minus(
    rc.Scalar(1),
    rc.Index(tokens_arr, I[:, ParenTokenizer.PAD_TOKEN]),
    name="pos_mask",
)
circuit = model_info.bind_to_input(circuit, tok_embeds, circ_dict["t.w.pos_embeds"], attn_mask)

# [markdown]
"""
Step 3: rewrites the circuit into a more conveninet structure to work with using `configure_transformer`.
This flattens out the residual stream (as opposed to the nested layer structure originally) and, pushes down the weight bindings, and separates out each attention layer into a sum of heads.
"""

circuit = circuit.update(
    "t.bind_w",
    lambda c: configure_transformer(
        c,
        To.ATTN_HEAD_MLP_NORM,
        split_by_head_config="full",
        use_pull_up_head_split=True,
        use_flatten_res=True,
        flatten_components=True,
    ),
)

# [markdown]
"""
Some additional misc rewrites.

We substitute the inputs to be duplicated everywhere they apperar in the model instead of being in one outer module bind.

We also index as we only care about the classification at position 0, and use `rc.conform_all_modules` to replace any remaining symbolic shapes with their numeric values.
"""

circuit = circuit.cast_module().substitute()
circuit = rc.Index(circuit, I[0]).rename("logits_pos0")
circuit = rc.conform_all_modules(circuit)

# %% [markdown]
"""
Finally, some custom renames that make the circuit more intuitive.
"""
circuit = circuit.update("t.call", lambda c: c.rename("logits"))
circuit = circuit.update("t.call", lambda c: c.rename("logits_with_bias"))
circuit = circuit.update(rc.Regex(r"[am]\d(.h\d)?$"), lambda c: c.rename(c.name + ".inner"))
circuit = circuit.update("t.inp_tok_pos", lambda c: c.rename("embeds"))
circuit = circuit.update("t.a.mask", lambda c: c.rename("padding_mask"))

for l in range(model_info.params.num_layers):
    circuit = circuit.update(f"b{l}.m", lambda c: c.rename(f"m{l}"))
    circuit = circuit.update(f"b{l}.a.h0", lambda c: c.rename(f"a{l}.h0"))
    circuit = circuit.update(f"b{l}.a.h1", lambda c: c.rename(f"a{l}.h1"))

    next = "final" if l == model_info.params.num_layers - 1 else f"a{l+1}"
    circuit = circuit.update(f"b{l}", lambda c: c.rename(f"{next}.input"))

printer = rc.PrintHtmlOptions(
    shape_only_when_necessary=False,
    traversal=rc.restrict(
        rc.IterativeMatcher("embeds", "padding_mask", "final.norm", rc.Regex(r"^[am]\d(.h\d)?$")), term_if_matches=True
    ),
)
# %%

if PRINT_CIRCUITS:
    printer.print(circuit)
    circuit_graph_ui(circuit, default_hidden=ui_hidden_matcher.get(circuit))


# %% [markdown]
"""
## dataset and experiment code
We have a custom dataset class that precomputes some features of paren sequences, and handles pretty printing / etc.
"""

ds = ParenDataset.load()


def bce_with_logits_loss(logits: torch.Tensor, labels: torch.Tensor):
    """
    Computes the binary cross entropy loss for the provided labels.
    logits: [batch, 2]. Class 0 is unbalanced logit, class 1 is balanced logit.
    labels: [batch]. True if balanced.
    """
    targets = labels.to(dtype=logits.dtype, device=logits.device)
    logit_diff = logits[..., 1] - logits[..., 0]
    correct = (logit_diff > 0) == targets
    loss = binary_cross_entropy_with_logits(logit_diff, targets, reduction="none")
    return loss, correct


def paren_experiment(
    circuit: rc.Circuit,
    dataset: ParenDataset,
    corr: Correspondence,
    checks: ExperimentCheck = DEFAULT_CHECKS,
    random_seed=1,
    actually_run=ACTUALLY_RUN,
    num_examples=NUM_EXAMPLES,
    batch_size=BATCH_SIZE,
    **kwargs,
) -> Tuple[ScrubbedExperiment, Optional[float]]:
    ex = Experiment(
        circuit,
        dataset,
        corr,
        random_seed=random_seed,
        check=checks,
        **kwargs,
    )
    scrubbed = ex.scrub(num_examples, treeify=actually_run)
    overall_loss: Optional[float] = None
    if actually_run:
        logits = scrubbed.evaluate(
            ExperimentEvalSettings(
                optim_settings=rc.OptimizationSettings(
                    max_memory=MAX_MEMORY,
                ),
                device_dtype=rc.TorchDeviceDtypeOp(device=EVAL_DEVICE),
                optimize=True,
                batch_size=batch_size,
            ),
        )
        ref_ds = ParenDataset.unwrap(scrubbed.ref_ds)
        labels = ref_ds.is_balanced.value

        def loss_str(mask):
            loss, correct = bce_with_logits_loss(logits[mask], labels[mask])
            loss = loss.cpu()
            std_err = loss.std() / len(loss) ** 0.5

            return f"{loss.mean():.3f}  SE={std_err:.3f}  acc={correct.float().mean():.1%} "

        print(f"  overall:               {loss_str(slice(None))}")
        print(f"    on bal:              {loss_str(labels.to(dtype=torch.bool))}")
        print(f"    on unbal:            {loss_str(~labels.to(dtype=torch.bool))}")
        print(f"    on count failures:   {loss_str(~ref_ds.count_test.to(dtype=torch.bool))}")  # type: ignore
        print(f"    on horizon failures: {loss_str(~ref_ds.horizon_test.to(dtype=torch.bool))}")  # type: ignore
        overall_loss = bce_with_logits_loss(logits, labels)[0].mean().item()

    return scrubbed, overall_loss


def check_loss(loss: Optional[float], target: float, std_err: float):
    # we usually set tol to be 4 * SE
    assert loss is not None
    err = abs(loss - target)
    assert err < 4 * std_err, f"err too large! loss ({loss:.2f}) != target ({loss:.2f}) ± 4*SE ({std_err:.2f})"
    if err > 2 * std_err:
        raise Warning("Err is kinda large! loss ({loss:.2f}) != target ({loss:.2f}) ± 2*SE ({std_err:.2f})")


# %% [markdown]
"""
Helpful tidbit on tests:
Most of the tests in this file raise assertion errors that contain extra data, for instance the objects from the comparison that failed. It can be convenient to catch this data to debug. For instance:

```
def check_eq(a, b):
    assert a == b, ("not equal!", a, b)

try:
    check_eq(0, 1)
except AssertionError as e:
    a, b = e.args[0][1], e.args[0][2]
    print(a, b)
```
"""


# %% [markdown]
"""
## Experiment 0
To start with let's measure two baselines:
  - running the model normally
  - interchanging the logits randomly

Make causal scrubbing experiments that impliment both of these. In each case there should be a single interp node named "logits".

The tests do explictly check that the interp nodes in the correspondence are named correctly, in order to facilitate more helpful feedback.
"""

if "SOLUTION":
    corr0a = Correspondence()
    corr0a.add(InterpNode(cond_sampler=ExactSampler(), name="logits"), corr_root_matcher)
else:
    corr0a = Correspondence()
if MAIN:
    tests.t_ex0a_corr(corr0a)
    print("\nEx0a: Exact sampling")
    ex0a, loss0a = paren_experiment(circuit, ds, corr0a)
    check_loss(loss0a, 0, 0.01)

if "SOLUTION":
    corr0b = Correspondence()
    corr0b.add(InterpNode(cond_sampler=UncondSampler(), name="logits"), corr_root_matcher)
else:
    corr0b = Correspondence()
if MAIN:
    tests.t_ex0b_corr(corr0b)
    print("\nEx0b: Interchanging logits")
    ex0b, loss0b = paren_experiment(circuit, ds, corr0b)
    check_loss(loss0b, 4.30, 0.12)

# %% [markdown]
"""
## Experiment 1
Now, let's construct a basic experiment to determine the role that different heads play.
We'll start by testing the following claimed hypothesis:
 - Heads 1.0 and 2.0 compute the count test, and check that there are equal numbers of open and close parentheses
 - Head 2.1 computes the horizon test.

### Matchers
"""
if "SOLUTION":
    # There are several ways to define this matcher, eg:
    # m_10 = rc.IterativeMatcher("final.input").chain(rc.restrict("a1.h0", end_depth=2))
    # or defining all_components = {"a0.h0", "a0.h1", "m0", ..., "m2"} and then
    # m_10 = chain_excluding(corr_root_matcher, "a1.h0", all_components - "a1.h0")
    # It's just down to personal preference / what is clearer in the circumstance.
    m_10 = chain_excluding(corr_root_matcher, "a1.h0", {"m2", "m1", "a2.h0", "a2.h1"})
    m_20 = chain_excluding(corr_root_matcher, "a2.h0", "m2")
    m_21 = chain_excluding(corr_root_matcher, "a2.h1", "m2")
else:
    """
    Define the following matchers. You only want to match _direct_ paths, that is paths through the
    residual stream and not through direct paths. This can be accomplished by calling `rc.restrict` or
    the chain_excluding utilty included in causal scrubbing code.
    """
    m_10 = rc.IterativeMatcher()
    m_20 = rc.IterativeMatcher()
    m_21 = rc.IterativeMatcher()

if MAIN:
    tests.t_m_10(m_10)
    tests.t_m_20(m_20)
    tests.t_m_21(m_21)

# %%  [markdown]
"""
### Cond Samplers 
"""


def passes_count(d: ParenDataset) -> torch.Tensor:
    """
    Returns a bool tensor of shape [len(d)]
    Result is true when the corresponding datum has equal numbers of open and close parens
    """
    # not used in solution, as implimented in ParenDataset
    raise NotImplementedError


def passes_horizon(d: ParenDataset) -> torch.Tensor:
    """
    Returns a bool tensor of shape [len(d)]
    Result is true when the corresponding datum passes the right to left horizon test as described in the [writeup](https://www.lesswrong.com/s/h95ayYYwMebGEYN5y/p/kjudfaQazMmC74SbF#Algorithm).
    """
    # not used in solution, as implimented in ParenDataset
    raise NotImplementedError


if "SOLUTION":
    count_cond = FuncSampler(lambda d: ParenDataset.unwrap(d).count_test)
    horizon_cond = FuncSampler(lambda d: ParenDataset.unwrap(d).horizon_test)
else:
    """
    Define the following cond samplers using FuncSamplers.

    Write your own functions for this based on `ParenDataset.tokens_flat`.

    (Yes, there are predefined properties on ParenDataset. You can use them if you are short on time and want to skip this test. They also use caching to make things faster, so if speed gets to be annoyingly slow for later experiments you could switch over).
    """
    count_cond = FuncSampler(lambda d: passes_count(ParenDataset.unwrap(d)))
    horizon_cond = FuncSampler(lambda d: passes_horizon(ParenDataset.unwrap(d)))

if MAIN:
    tests.t_count_cond(count_cond)
    tests.t_horizon_cond(horizon_cond)

# %% [markdown]
"""
This first correspondence should have 4 nodes:
 - The root node, named "logits", with an ExactSampler (any sampler that agrees on the labels will be equivilant,
 but an exact sampler is somewhat more computationally efficient).
 - Three nodes for the three heads of interest, named "10", "20", and "21". The first two should use the cond sampler provided (`count_cond` for now), the third should use the `horizon_cond` sampler.

The exact interp node names are checked by the test, which allows it to give more meaningful feedback. 
"""


def make_ex1_corr(cs_for_h10_and_h20) -> Correspondence:
    """SOLUTION"""
    corr = Correspondence()
    i_logits = InterpNode(name="logits", cond_sampler=ExactSampler())
    corr.add(i_logits, corr_root_matcher)
    corr.add(i_logits.make_descendant(cs_for_h10_and_h20, "10"), m_10)
    corr.add(i_logits.make_descendant(cs_for_h10_and_h20, "20"), m_20)
    corr.add(i_logits.make_descendant(horizon_cond, "21"), m_21)
    return corr


# %%
if MAIN:
    print("\nEx1a: Just the count cond")
    tests.t_ex1_corr(make_ex1_corr, count_cond)
    ex1a, loss1a = paren_experiment(circuit, ds, make_ex1_corr(count_cond))
    check_loss(loss1a, 0.52, 0.04)

    if PRINT_CIRCUITS:
        ex1a.print()


# %% [markdown]
"""
As discussed in the writeup, we can more accurately capture the equivilance classes of 1.0 and 2.0's output by including if the first parenthesis is open or closed.

This is a natural feature for these heads to use: a sequence will always be unbalanced if it starts with a close parenthesis, and as these heads depend strongly on the residual stream at position 1 anyhow (as we will show in experiement 2) the information is readily accessible.

(reminder: position 0 is always the [START] token, so position 1 is the first parentheses. All sequences in our dataset have >= 2 parentheses in them, so you can assume position 1 is either an open or close paren.)

Define some new cond samplers that incorporate this feature.
"""


def passes_starts_open(d: ParenDataset) -> torch.Tensor:
    """
    Returns a bool tensor of shape [len(d)].
    Result is true when the corresponding datum starts with '('.
    """
    raise NotImplementedError


def passes_count_open(d: ParenDataset) -> torch.Tensor:
    """
    Returns a bool tensor of shape [len(d)].
    Result is true when the corresponding datum starts with '(' and there are equal numbers of open and close parens in the entire sequence.
    """
    raise NotImplementedError


if "SOLUTION":
    start_open_cond = FuncSampler(lambda d: ParenDataset.unwrap(d).starts_with_open)
    count_open_cond = FuncSampler(lambda d: ParenDataset.unwrap(d).count_test & ParenDataset.unwrap(d).starts_with_open)
else:
    """
    Two more samplers! the first that checks that the first paren is an open parentheses,
    the next tests the input passes count test AND the first paren is open.

    We don't use the pure start_open test yet, but we will soon and it's nice to define it here.
    """
    start_open_cond = FuncSampler(lambda d: passes_starts_open(ParenDataset.unwrap(d)))
    count_open_cond = FuncSampler(lambda d: passes_count_open(ParenDataset.unwrap(d)))

if MAIN:
    tests.t_start_open_cond(start_open_cond)
    tests.t_count_open_cond(count_open_cond)


# %%
if MAIN:
    print("\nEx1b: Without a0")
    tests.t_ex1_corr(make_ex1_corr, count_open_cond)
    ex1b, loss1b = paren_experiment(circuit, ds, make_ex1_corr(count_open_cond))
    check_loss(loss1b, 0.30, 0.04)

    if PRINT_CIRCUITS:
        ex1b.print()

# [markdown]
"""
Bonus: Can you improve on the loss by specifying other direct paths, or choosing better features to ensure
agreement along?
"""
# %%
if "SOLUTION":
    if MAIN:
        print("\nEx1 bonus 1: With a0")
        corr = make_ex1_corr(count_open_cond)

        i_00 = corr.get_by_name("logits").make_descendant(count_open_cond, "00")
        m_00 = chain_excluding(corr_root_matcher, "a0.h0", {"m2", "m1", "m0", "a2.h0", "a2.h1", "a1.h0", "a1.h1"})

        corr.add(i_00, m_00)
        ex1_bonus1 = paren_experiment(circuit, ds, corr)


# %%
# We only test up until this point in CircleCI so that it's fast
if "SKIP":
    NotebookInTesting.exit_if_in_testing()

# %% [markdown]
"""
## Experiment 2! Diving into heads 1.0 and 2.0
We are going split up experiment 2 into four parts:
 - Part 1: 1.0 and 2.0 only depend on their input at position 1
 - Part 2 (ex2a in writeup): 1.0 and 2.0 only depend on:
    - the output of 0.0 (which computes $p$, the proportion of open parentheses) and
    - the embeds (which encode if the first paren is open)  
 - Part 3: Projecting the output of 0.0 onto a single direction
 - Part 4 (ex2b in writeup): Estimate the output of 0.0 with a function $\phi(p)$

## Part 1: Splitting up the input to 1.0 and 2.0 by sequence position
One of the claims we'd like to test is that only the input at position 1 (the first paren position) matters for both heads 1.0 and 2.0.
Currently, however, there is no node of our circuit corresponding to "the input at position 1". Let's change that!

Write a `separate_pos1` function that will transform a circuit `node` into:
```
'node_concat' Concat
  'node.pos_0' Index [0:1, :]
    'node'
  'node.pos_1' Index [1:2, :]
    'node'
  'node.pos_2_41' Index [2:42, :]
    'node'
```
This can be acomplished by calling `split_to_concat()` (from algebraic_rewrites.py) and `.rename`-ing the result.

Then split the input node, but only along paths that are reached through head 2.0 and 1.0. (We don't want to split the input to 2.1 in particular, as we'll split that differently later.)

If you are struggling to get the exact correct circuit, you are free to import it from the solution file and print it out. You can also try `print(rc.diff_circuits(your_circuit, our_circuit))` though

Yes, the tests are very strict about naming things exactly correctly.
This is partially because it is convenient for tests, but also because names are really important!
Good names reduce confusion about what that random node of the circuit actually means.
Mis-naming nodes is also a frequent cause of bugs, e.g. a matcher that traverses a path that it wasn't supposted to.
"""

# %%
def separate_pos1(c: rc.Circuit) -> rc.Circuit:
    "SOLUTION"
    return split_to_concat(
        c, 0, [0, 1, torch.arange(2, 42)], partitioning_idx_names=["pos0", "pos1", "pos2_42"], use_dot=True
    ).rename(f"{c.name}_concat")


ex2_part1_circuit = circuit
if "SOLUTION":
    ex2_part1_circuit = ex2_part1_circuit.update(rc.IterativeMatcher("a2.h0").chain("a2.input"), separate_pos1)
    ex2_part1_circuit = ex2_part1_circuit.update(rc.IterativeMatcher("a1.h0").chain("a1.input"), separate_pos1)

if MAIN and PRINT_CIRCUITS:
    subcirc = ex2_part1_circuit.get_unique(rc.IterativeMatcher("a2.h0").chain("a2.input_concat"))
    printer.print(subcirc)
    circuit_graph_ui(subcirc, default_hidden=ui_hidden_matcher.get(subcirc))


if MAIN:
    tests.t_ex2_part1_circuit(ex2_part1_circuit)

# %% [markdown]
"""
Now we can test the claim that both 1.0 and 2.0 only cares about positon 1!

We'll need new matchers, which just matches the pos_1 input.
"""

if "SOLUTION":
    m_10_p1 = m_10.chain("a1.input.at_pos1")
    m_20_p1 = m_20.chain("a2.input.at_pos1")
else:
    m_10_p1 = rc.IterativeMatcher()
    m_20_p1 = rc.IterativeMatcher()

if MAIN:
    tests.t_m_10_p1(m_10_p1)
    tests.t_m_20_p1(m_20_p1)

# %% [markdown]
"""
Then create a correspondence that extends the one returned by `make_ex1_corr(count_open_cond)` so that both 1.0 and 2.0 only use information from position 1. `Correspondence.get_by_name` is useful here.

Have your new nodes be named "10_p1" and "20_p1".
"""


def make_ex2_part1_corr() -> Correspondence:
    "SOLUTION"
    corr = make_ex1_corr(count_open_cond)
    i_10_p1 = corr.get_by_name("10").make_descendant(name="10_p1", cond_sampler=count_open_cond)
    corr.add(i_10_p1, m_10_p1)
    i_20_p1 = corr.get_by_name("20").make_descendant(name="20_p1", cond_sampler=count_open_cond)
    corr.add(i_20_p1, m_20_p1)
    return corr


if MAIN:
    tests.t_make_ex2_part1_corr(make_ex2_part1_corr())
    print("\nEx 2 part 1: 1.0/2.0 depend on position 1 input")
    ex2_p1, loss2_p1 = paren_experiment(ex2_part1_circuit, ds, make_ex2_part1_corr())


# %% [markdown]
"""
### Part 2
We now construct experiment 2a from the writeup. We will be strict about where 1.0 and 2.0 learn the features they depend on. We claim that the 'count test' is determined by head 0.0 checking the exact proportion of open parens in the sequence and outputting this into the residual stream at position 1.

We thus need to also split up the output of attention head 0.0, so we can specify it only cares about the output of this head at position 1. Again, let's only split it for the branch of the circuit we are working with: copies of 0.0 that are upstream of either `m_10_p1` or `m_20_p1`. 
"""

ex2_part2_circuit = ex2_part1_circuit
if "SOLUTION":
    ex2_part2_circuit = ex2_part2_circuit.update((m_10_p1 | m_20_p1).chain("a0.h0"), separate_pos1)

if MAIN and PRINT_CIRCUITS:
    printer.print(ex2_part2_circuit.get_unique(m_10_p1))

if MAIN:
    tests.t_ex2_part2_circuit(ex2_part2_circuit)

# %% [markdown]
"""
First, make a new cond sampler that samples an input that agrees on what is called $p_1^($ in the writeup. This can be done with a FuncSampler based on a function with the following equivalence classes:
 - one class for _all_ inputs that start with a close parenthesis
 - one class for every value of $p$ (proportion of open parentheses in the entire sequence)

Note the actual values returned aren't important, just the equivialance clases.
"""


def p1_if_starts_open(d: ParenDataset):
    """Returns a tensor of size [len(ds)]. The value represents p_1 if the sequence starts open, and is constant otherwise"""
    "SOLUTION"
    return torch.where(
        d.starts_with_open,
        d.p_open_after[:, 1],
        torch.tensor(-1.0, dtype=torch.float32),
    )


p1_open_cond = FuncSampler(lambda d: p1_if_starts_open(ParenDataset.unwrap(d)))

# %% [markdown]
"""And some matchers"""

if "SOLUTION":
    m_10_p1_h00 = m_10_p1.chain("a0.h0.at_pos1")
    m_20_p1_h00 = m_20_p1.chain("a0.h0.at_pos1")
else:
    m_10_p1_h00 = rc.IterativeMatcher()
    m_20_p1_h00 = rc.IterativeMatcher()

# %% [markdown]
"""
Now make the correspondence!

You should add 4 nodes to the correspondence from part 1:
 - "10_p1_00"
 - "20_p1_00"
 - "10_p1_emb"
 - "20_p1_emb"

"""


def make_ex2_part2_corr() -> Correspondence:
    "SOLUTION"
    corr = make_ex2_part1_corr()

    i_10_p1 = corr.get_by_name("10_p1")
    i_20_p1 = corr.get_by_name("20_p1")

    # indirect paths to 0.0
    i_10_p1_00 = i_10_p1.make_descendant(name="10_p1_00", cond_sampler=p1_open_cond)
    i_20_p1_00 = i_20_p1.make_descendant(name="20_p1_00", cond_sampler=p1_open_cond)

    corr.add(i_10_p1_00, m_10_p1_h00)
    corr.add(i_20_p1_00, m_20_p1_h00)

    i_10_p1_emb = i_10_p1.make_descendant(name="10_p1_emb", cond_sampler=start_open_cond)
    i_20_p1_emb = i_20_p1.make_descendant(name="20_p1_emb", cond_sampler=start_open_cond)

    corr.add(i_10_p1_emb, chain_excluding(m_10_p1, "embeds", "a0.h0"))
    corr.add(i_20_p1_emb, chain_excluding(m_20_p1, "embeds", "a0.h0"))

    return corr


if MAIN:
    tests.t_ex2_part2_corr(make_ex2_part2_corr())
    print("\nEx 2 part 2 (2a in writeup): 1.0/2.0 depend on position 0.0 and emb")
    ex2a, loss2a = paren_experiment(ex2_part2_circuit, ds, make_ex2_part2_corr())
    check_loss(loss2a, 0.55, 0.04)


# %% [markdown]
"""
### Part 3: Projecting 0.0 onto a single direction
#### Circuit rewrite
Another claim we would like to test is that only the output of 0.0 written in a particular direction is important.

To do this we will rewrite the output of 0.0 as the sum of two terms: the (projection)[https://en.wikipedia.org/wiki/Vector_projection] and rejection (aka the perpendicular component) along this direction.
"""

# %%

h00_open_vector = get_h00_open_vector(MODEL_ID)


def project_into_direction(c: rc.Circuit, v: torch.Tensor = h00_open_vector) -> rc.Circuit:
    """
    Return a circuit that computes `c`: [seq_len, 56] projected the direction of vector `v`: [56].
    Call the resulting circuit `{c.name}_projected`.
    """
    "SOLUTION"
    v_dir_array = rc.Array(v / torch.linalg.norm(v), "dir")
    return rc.Einsum.from_einsum_string(
        "s e, e, f -> s f", c.rename(c.name + "_orig"), v_dir_array, v_dir_array, name=f"{c.name}_projected"
    )


if MAIN:
    tests.t_project_into_direction(project_into_direction)


def get_ex2_part3_circuit(c: rc.Circuit, project_fn=project_into_direction):
    """
    Uses `residual_rewrite` to write head 0.0 at position 1 (when reached by either `m_10_p1_h00` or `m_20_p1_h00`), as a sum of the projection and the rejection along h00_open_vector. The head retains it's same name, with children named `{head.name}_projected` and `{head.name}_projected_residual`.
    """
    "SOLUTION"
    split = lambda h00: residual_rewrite(h00, project_fn(h00), "projected")[0]
    return c.update((m_10_p1_h00 | m_20_p1_h00).chain("a0.h0"), split)


ex2_part3_circuit = get_ex2_part3_circuit(ex2_part2_circuit)

if MAIN and PRINT_CIRCUITS:
    proj_printer = printer.evolve(
        traversal=rc.new_traversal(term_early_at={"a0.h0.at_pos0", "a0.h0.at_pos2_42", "a0.h0_orig"})
    )
    subcirc = ex2_part3_circuit.get_unique(m_10_p1.chain("a0.h0_concat"))
    proj_printer.print(subcirc)

if MAIN:
    tests.t_ex2_part3_circuit(get_ex2_part3_circuit)

# %% [markdown]
"""
Now make the correspondence. Be sure to avoid the residual node!

This correspondence requires adding two new nodes:
 - "10_p1_00_projected"
 - "20_p1_00_projected"
"""


def make_ex2_part3_corr() -> Correspondence:
    "SOLUTION"
    corr = make_ex2_part2_corr()
    corr.add(
        corr.get_by_name("10_p1_00").make_descendant(p1_open_cond, "10_p1_00_projected"),
        chain_excluding(m_10_p1_h00, "a0.h0_projected", "a0.h0_projected_residual"),
    )
    corr.add(
        corr.get_by_name("20_p1_00").make_descendant(p1_open_cond, "20_p1_00_projected"),
        chain_excluding(m_20_p1_h00, "a0.h0_projected", "a0.h0_projected_residual"),
    )
    return corr


if MAIN:
    tests.t_ex2_part3_corr(make_ex2_part3_corr())
    print("\nEx 2 part 3: Projecting h00 into one direction")
    ex2_p3, loss2_p3 = paren_experiment(ex2_part3_circuit, ds, make_ex2_part3_corr())


# %% [markdown]
"""
### Part 4: The $\phi$ function
"""
# %%
def compute_phi_circuit(tokens: rc.Circuit):
    """
    tokens: [seq_len, vocab_size] array of one hot tokens representing a sequence of parens
    (see ParenTokenizer for the one_hot ordering)

    Returns a circuit that computes phi: tokens -> R^56
    phi = h00_open_vector(2p - 1)
    where p = proportion of parens in `tokens` that are open.

    Returns a circuit with name 'a0.h0_phi'.
    """
    "SOLUTION"
    num_opens = rc.Index(tokens, [slice(None), ParenTokenizer.OPEN_TOKEN], name="is_open").sum(
        axis=-1, name="num_opens"
    )
    num_closes = rc.Index(tokens, [slice(None), ParenTokenizer.CLOSE_TOKEN], name="is_close").sum(
        axis=-1, name="num_closes"
    )
    num_parens = rc.Add(num_opens, num_closes, name="num_parens")
    p_open = num_opens.mul(rc.reciprocal(num_parens), name="p_open")
    p_open_mul_factor = p_open.mul_scalar(2).sub(rc.Scalar(1), name="p_open_mul_factor")
    dir = rc.Array(h00_open_vector, "open direction")
    return dir.mul(p_open_mul_factor, name="a0.h0_phi")


if MAIN:
    tests.t_compute_phi_circuit(compute_phi_circuit)

# %%


def get_ex2_part4_circuit(orig_circuit: rc.Circuit = ex2_part2_circuit, compute_phi_circuit_fn=compute_phi_circuit):
    """
    Split the output of head 0.0 at position 1, when reached through the appropriate paths, into a phi estimate
    and the residual of this estimate.

    The resulting subcircuit should have name 'a0.h0' with children 'a0.h0_phi' and 'a0.h0_phi_residual'.
    """
    "SOLUTION"
    split_by_phi = lambda c: residual_rewrite(c, compute_phi_circuit_fn(c.get_unique("tokens")), "phi")[0]
    return orig_circuit.update((m_10_p1_h00 | m_20_p1_h00).chain("a0.h0"), split_by_phi)


ex2_part4_circuit = get_ex2_part4_circuit()

if MAIN and PRINT_CIRCUITS:
    proj_printer = printer.evolve(
        traversal=rc.new_traversal(term_early_at={"a0.h0.at_pos0", "a0.h0.at_pos2_42", "a0.h0_orig"})
    )
    subcirc = ex2_part4_circuit.get_unique(m_10_p1.chain("a0.h0_concat"))
    proj_printer.print(subcirc)


if MAIN:
    tests.t_ex2_part4_circuit(get_ex2_part4_circuit)

# %% [markdown]
"""
And now make the correspondence -- it should be very similar to the one from part 3. Build on top of the one from part **2**, with new node names "10_p1_00_phi" and "20_p1_00_phi".
"""


def make_ex2_part4_corr() -> Correspondence:
    "SOLUTION"
    corr = make_ex2_part2_corr()
    corr.add(
        corr.get_by_name("10_p1_00").make_descendant(p1_open_cond, "10_p1_00_phi"),
        chain_excluding(m_10_p1_h00, "a0.h0_phi", "a0.h0_phi_residual"),
    )
    corr.add(
        corr.get_by_name("20_p1_00").make_descendant(p1_open_cond, "20_p1_00_phi"),
        chain_excluding(m_20_p1_h00, "a0.h0_phi", "a0.h0_phi_residual"),
    )
    return corr


if MAIN:
    tests.t_ex2_part4_corr(make_ex2_part4_corr())
    print("Ex2 part 4 (2b in writeup): replace a0 by phi(p)")
    ex2b, loss2b = paren_experiment(ex2_part4_circuit, ds, make_ex2_part4_corr())
    check_loss(loss2b, 0.53, 0.04)

# %% [markdown]
"""
Congradulations! This is the end of the main part of today's content. Below is some additional content that covers experiments 3 and 4 from the writeup, although with less detailed testing and instructions.
"""

# %%
if "SKIP":
    ######### Bonus experiments! ########
    # some bonus tests that don't appear in the writeup
    # these test the direct dependances of how information flows from a0 -> 2.0
    # note we don't have a dependance on the tokens here for the 'starts open'
    # which is probably a shortcoming of these hypotheses

    def make_ex2_explicit_paths_corr(from_10=True, to_m0_m1_a0=True, all_to_a0=True) -> Correspondence:
        corr = make_ex1_corr(count_open_cond)
        i_20 = corr.get_by_name("20")
        i_20_p1 = i_20.make_descendant(name="20_p1", cond_sampler=count_open_cond)
        corr.add(i_20_p1, m_20_p1)

        if to_m0_m1_a0:
            i_20_p1_a0 = i_20_p1.make_descendant(name="20_p1_a0", cond_sampler=count_open_cond)
            i_20_p1_m0 = i_20_p1.make_descendant(name="20_p1_m0", cond_sampler=count_open_cond)
            i_20_p1_m1 = i_20_p1.make_descendant(name="20_p1_m1", cond_sampler=count_open_cond)

            m_20_p1_m0 = chain_excluding(m_20_p1, "m0", "m1")
            m_20_p1_m1 = m_20_p1.chain("m1")
            corr.add(i_20_p1_a0, chain_excluding(m_20_p1, "a0.h0", {"m0", "a1.h0", "a1.h1", "m1"}))
            corr.add(i_20_p1_m0, m_20_p1_m0)
            corr.add(i_20_p1_m1, m_20_p1_m1)

            if all_to_a0:
                # m0 depends on a0
                i_20_p1_m0_a0 = i_20_p1_m0.make_descendant(name="20_p1_m0_00", cond_sampler=p1_open_cond)

                # m1 depends on a0, m0
                i_20_p1_m1_a0 = i_20_p1_m1.make_descendant(name="20_p1_m1_00", cond_sampler=p1_open_cond)
                i_20_p1_m1_m0 = i_20_p1_m1.make_descendant(name="20_p1_m1_m0", cond_sampler=p1_open_cond)
                # m1.m0 depends on a0
                i_20_p1_m1_m0_a0 = i_20_p1_m1_m0.make_descendant(name="20_p1_m1_m0_00", cond_sampler=p1_open_cond)

                corr.add(i_20_p1_m0_a0, m_20_p1_m0.chain("a0.h0"))
                corr.add(i_20_p1_m1_a0, chain_excluding(m_20_p1_m1, "a0.h0", "m0"))
                corr.add(i_20_p1_m1_m0, m_20_p1_m1.chain("m0"))
                corr.add(i_20_p1_m1_m0_a0, m_20_p1_m1.chain("m0").chain("a0.h0"))

        if from_10:
            i_10 = corr.get_by_name("10")
            i_10_p1 = i_10.make_descendant(name="10_p1", cond_sampler=count_open_cond)
            corr.add(i_10_p1, m_10_p1)

            if to_m0_m1_a0:
                i_10_p1_a0 = i_10_p1.make_descendant(name="10_p1_a0", cond_sampler=count_open_cond)
                i_10_p1_m0 = i_10_p1.make_descendant(name="10_p1_m0", cond_sampler=count_open_cond)

                corr.add(i_10_p1_a0, chain_excluding(m_10_p1, "a0.h0", {"m0"}))
                m10_p1_m0 = m_10_p1.chain("m0")
                corr.add(i_10_p1_m0, m10_p1_m0)

                if all_to_a0:
                    i_10_p1_m0_a0 = i_10_p1_m0.make_descendant(name="10_p1_m0_00", cond_sampler=p1_open_cond)
                    corr.add(i_10_p1_m0_a0, m10_p1_m0.chain("a0.h0"))

        return corr

    #%%
    if MAIN:
        print("Ex 2_bonus_a: 2.0 depends on only pos1")
        ex2_bonus_a = paren_experiment(ex2_part2_circuit, ds, make_ex2_explicit_paths_corr(False, False, False))

        print("Ex 2_bonus_b: 2.0 depends on only pos1, which depends on a0 + m0 + m1")
        ex2_bonus_b = paren_experiment(ex2_part2_circuit, ds, make_ex2_explicit_paths_corr(False, True, False))

        print("Ex 2_bonus_c: 2.0 depends on only pos1, which depends on a0 + m0 + m1, which bottom out at a0")
        ex2_bonus_c = paren_experiment(ex2_part2_circuit, ds, make_ex2_explicit_paths_corr(False, True, True))

        if PRINT_CIRCUITS:
            ex2_part2_circuit.print(printer)

# %%
"""
# Experiment 3
"""


def separate_all_seqpos(c: rc.Circuit) -> rc.Circuit:
    """
    Separate c into all possible sequence positions.
    c is renamed to `{c.name}_concat`, with children `{c.name}.at_pos{i}`
    """
    "SOLUTION"
    return split_to_concat(c, 0, range(42), partitioning_idx_names=[f"pos{i}" for i in range(42)], use_dot=True).rename(
        f"{c.name}_concat"
    )


if MAIN:
    tests.t_separate_all_seqpos(separate_all_seqpos)

ex3_circuit = circuit
if "SOLUTION":
    ex3_circuit = ex3_circuit.update(rc.IterativeMatcher("a2.h1").chain("a2.input"), separate_all_seqpos)

if MAIN and PRINT_CIRCUITS:
    printer.print(ex3_circuit.get_unique(rc.IterativeMatcher("a2.input_concat")))

if MAIN:
    tests.t_ex3_circuit(ex3_circuit)


# %% [markdown]
"""
When adjusted = True, use the `ds.adjusted_p_open_after` attribute instead of `ds.p_open_after` to compute the horizon test.

One possible gotcha in this section is late-binding-closures messing with the values of i. I think if you follow the outline you should be fine, but if you get strange bugs it's one possibility.
"""


def to_horizon_vals(d: ParenDataset, i: int, adjusted: bool = False) -> torch.Tensor:
    """
    Returns a value for the horizon_i test dividing up the input datums into 5 equivalence classes.
    The actual numerical return values don't have inherent meaning, but are defined as follows:
        0 on padding,
        positive on plausibly-balanced positions,
        negative on unbalance-evidence positions,
        1 / -1 on END_TOKENS,
        2 / -2 on non-end tokens
    """
    "SOLUTION"
    ps: torch.Tensor = d.adjusted_p_open_after if adjusted else d.p_open_after
    toks = d.tokens_flat.value
    assert toks.shape == ps.shape
    conds = [
        toks[:, i] == ParenTokenizer.PAD_TOKEN,
        (toks[:, i] == ParenTokenizer.END_TOKEN) & (toks[:, i - 1] == ParenTokenizer.OPEN_TOKEN),
        (toks[:, i] == ParenTokenizer.END_TOKEN) & (toks[:, i - 1] == ParenTokenizer.CLOSE_TOKEN),
        ps[:, i] > 0.5,
        ps[:, i] <= 0.5,
    ]
    vals = np.select(conds, [0, -1, 1, -2, 2], default=np.nan)
    if np.any(np.isnan(vals)):
        print(i)
        print(d[np.isnan(vals)][:10])
        print(ps[np.isnan(vals)][:10])
        raise AssertionError
    return torch.tensor(vals)


if MAIN:
    tests.t_to_horizon_vals(to_horizon_vals)


def get_horizon_cond(i: int, adjusted: bool) -> FuncSampler:
    """Func sampler for horizon_i"""
    "SOLUTION"
    return FuncSampler(lambda d: to_horizon_vals(ParenDataset.unwrap(d), i, adjusted))


def get_horizon_all_cond(adjusted: bool) -> FuncSampler:
    """Func sampler horizon_all"""
    "SOLUTION"

    def to_horizon_all_vals(d: ParenDataset):
        stacked = torch.stack([to_horizon_vals(d, i, adjusted) for i in range(0, 42)], dim=-1)
        return (stacked >= 0).all(dim=-1)

    return FuncSampler(lambda d: to_horizon_all_vals(ParenDataset.unwrap(d)))


if MAIN:
    tests.t_get_horizon_all_cond(get_horizon_all_cond)


def make_ex3_corr(adjusted: bool = False, corr=None) -> Correspondence:
    """
    `adjusted`: uses `adjusted_p_open_after` based conditions if True, `p_open_after` otherwise.
    `corr`: The starting corr. Uses experiemnt 1b corr by default.

    Makes the following modifications:
     - Changes the cond sampler on node `21` to be the horizon_all cond sampler.
     - Adds one node for each sequence position, called `21_p{i}` with horizon_i cond sampler.
     - Also adds a node `pos_mask`, ensuring the pos_mask of head 2.1 is sampled from an input with the same input length.
    """
    "SOLUTION"
    corr = op.unwrap_or(corr, make_ex1_corr(count_open_cond))

    i_21 = corr.get_by_name("21")
    i_21.cond_sampler = get_horizon_all_cond(adjusted)
    for i in range(0, 42):
        i_pos_i = i_21.make_descendant(name=f"21_p{i}", cond_sampler=get_horizon_cond(i, adjusted))
        corr.add(i_pos_i, m_21.chain(f"a2.input.at_pos{i}"))

    i_pos_mask = i_21.make_descendant(
        name="pos_mask", cond_sampler=FuncSampler(lambda d: ParenDataset.unwrap(d).input_lengths)
    )

    m_pos_mask = chain_excluding(m_21, "pos_mask", "a2.input")
    corr.add(i_pos_mask, m_pos_mask)

    return corr


# %% [markdown]
"""
Note, we have a mini-replication crisis and with the current code can't replicate the exact numbers from the writeup. The handling of the position mask is somewhat different, although much more sane imo. I haven't had time to diagnose the exact cause of the difference.

In any case, expect your loss to be closer to ~1.23 for ex3a and ~1.17 for ex3b
"""
if MAIN:
    print("splitting up 2.1 input by seqpos")

    tests.t_make_ex3_corr(make_ex3_corr)
    print("\nEx3a: first with real open proprotion")
    ex3a, loss3a = paren_experiment(ex3_circuit, ds, make_ex3_corr(adjusted=False))
    check_loss(loss3a, 1.124, 0.1)  # sad sad sad

    print("\nEx3b: now with adjusted open proportion")
    ex3b, loss3b = paren_experiment(ex3_circuit, ds, make_ex3_corr(adjusted=True))
    check_loss(loss3b, 1.140, 0.1)  # sad sad sad

    if PRINT_CIRCUITS:
        ex3b.print()

# %%

if "SKIP":

    def make_ex3c_corr(adjusted: bool):
        corr = make_ex3_corr(adjusted)
        for i in range(42):
            i_pi = corr.get_by_name(f"21_p{i}")
            m_pi = corr[i_pi]
            corr.add(
                i_pi.make_descendant(get_horizon_cond(i, adjusted), f"21_p{i}_a0"),
                chain_excluding(m_pi, "a0.h0", {"m0", "m1"}),
            )
            corr.add(
                i_pi.make_descendant(get_horizon_cond(i, adjusted), f"21_p{i}_m0"), chain_excluding(m_pi, "m0", "m1")
            )
            corr.add(i_pi.make_descendant(get_horizon_cond(i, adjusted), f"21_p{i}_m1"), m_pi.chain("m1"))

        return corr

    if MAIN and SLOW_EXPERIMENTS:
        print("Ex3c: Extending each seqpos to {a0.0, m0, m1}")
        ex3c = paren_experiment(ex3_circuit, ds, make_ex3c_corr(True))

    def make_ex3d_corr(adjusted: bool):
        corr = make_ex3_corr(adjusted)
        for i in range(42):
            i_pi = corr.get_by_name(f"21_p{i}")
            m_pi = corr[i_pi]

            corr.add(i_pi.make_descendant(ExactSampler(), f"21_p{i}_a0"), chain_excluding(m_pi, "a0.h0", {"m0", "m1"}))
            corr.add(i_pi.make_descendant(ExactSampler(), f"21_p{i}_m0"), chain_excluding(m_pi, "m0", "m1"))
            corr.add(i_pi.make_descendant(ExactSampler(), f"21_p{i}_m1"), m_pi.chain("m1"))

        return corr

    if MAIN and SLOW_EXPERIMENTS:
        print("Ex3d: Extending each seqpos to a0.0+m0+m1")
        ex3d = paren_experiment(ex3_circuit, ds, make_ex3d_corr(True))


# %% [markdown]
"""
Now, combine experiments 2 (phi rewrite) and 3 (with adj. proportion)! Expected loss is ~1.64
"""

ex4_circuit = circuit
if "SOLUTION":
    ex4_circuit = ex2_part4_circuit.update(rc.IterativeMatcher("a2.h1").chain("a2.input"), separate_all_seqpos)
    ex4_corr = make_ex3_corr(adjusted=True, corr=make_ex2_part4_corr())
else:
    ex4_corr = Correspondence()

if MAIN and PRINT_CIRCUITS:
    printer.print(ex4_circuit)

if MAIN:
    print("\nEx4: Ex2b (1.0 and 2.0 phi rewrite) + Ex3b (2.1 split by seqpos with p_adj)")
    ex4, loss4 = paren_experiment(ex4_circuit, ds, ex4_corr)
    check_loss(loss4, 1.7, 0.1)  # sad sad sad

# %%



# %%
import jax
import matplotlib.pyplot as plt
import numpy as np
import torch

import interp.tools.optional as op
from interp.model.gpt_model import Gpt
from interp.model.model_loading import load_model_mem_cached
from interp.tools.log import Idxs, KeyIdxs, LoggerCache, LogInfo, construct_mut_log_cache

from .setup import ParenDataset, ParenTokenizer

jax.config.update("jax_platform_name", "cpu")


#%% [markdown]
"""This notebook produces plots for the causal scrubbing paren balancer writeup."""
# %%
# define model

MODEL_ID = "jun9_paren_balancer"

ds = ParenDataset.load(MODEL_ID)
ds = ds[:1000]

# %%
def bce_with_logits_loss(logits, labels):
    targets = torch.tensor(labels, dtype=torch.float, device="cuda:0")
    logit_diff = logits[:, 1] - logits[:, 0]
    correct = (logit_diff > 0) == targets
    return torch.nn.BCEWithLogitsLoss(reduction="none")(logit_diff, targets), correct


def get_bound_jax_model(model_id: str):
    jax_model, jax_params, _ = load_model_mem_cached(model_id)
    return jax_model.bind(jax_params)


def run_dataset_on_jax_model(dataset: ParenDataset, model_id: str, log_info=None):
    mask = (dataset.tokens_flat.value == ParenTokenizer.PAD_TOKEN).cpu().numpy()
    config = Gpt.CallConfig(pos_mask=mask)
    log = construct_mut_log_cache(log_info, None)
    jax_model = get_bound_jax_model(model_id)
    out = jax_model(dataset.tokens_flat.value.cpu().numpy(), log=log, config=config)[:, 0, :]
    probs = jax.nn.softmax(out)
    return out, probs, op.map(log, lambda log: log.cache)


# %%
a2_by_head = KeyIdxs("blocks.attention.out_by_head", idxs=Idxs.single(2))
norm_in = KeyIdxs("final_out.norm.inp")
logger = LoggerCache.from_key_idxs([a2_by_head, norm_in])
out, probs, cache = run_dataset_on_jax_model(ds, MODEL_ID, log_info=LogInfo(logger))
cache = op.unwrap(cache)

h20_out = cache.get(a2_by_head)[:, 0, 0, :]
h21_out = cache.get(a2_by_head)[:, 1, 0, :]
all_terms = cache.get(norm_in)[:, 0, :]

# %%
def attribution_score(head_term, full_sum):
    assert head_term.ndim == 2 and full_sum.ndim == 2  # [batch, hiddendim]
    rem = full_sum - head_term  # []
    possible_sums = head_term[:, None, :] + rem[None, :, :]  # [batch.head, batch.remainders, hidden]
    logits = np.array(get_bound_jax_model(MODEL_ID).out(possible_sums))
    logit_diffs = logits[..., 1] - logits[..., 0]
    return logit_diffs.mean(1) - logit_diffs.mean()


h20_attr = attribution_score(h20_out, all_terms)
h21_attr = attribution_score(h21_out, all_terms)

# %%
# set some globals that later cells can change
# thanks late binding closures!
alpha = 0.5
size = 25


def mk_scatter(filter, label=None, color=None, ax=None):
    ax = plt.gca() if ax is None else ax
    ax.scatter(h20_attr[filter], h21_attr[filter], label=label, color=color, s=size, alpha=alpha)


is_balanced = np.array(ds.is_balanced.value, dtype="bool")
count_test = ds.count_test.bool()
horizon_test = ds.horizon_test.bool()
ele_and_open = count_test & horizon_test

# %%
mk_scatter(is_balanced, "balanced", "#1b9e77")
mk_scatter(count_test & ~horizon_test, "just horizon failure", "#e7298a")
mk_scatter(~count_test & horizon_test, "just count ${}^($ failure", "#d95f02")
mk_scatter(~count_test & ~horizon_test, "both failures", "#7570b3")

plt.xlabel("Logit difference from 2.0")
plt.ylabel("Logit difference from 2.1")
plt.legend()
plt.gcf().set_size_inches(5, 5)

plt.show()

# %%
ele_and_open = count_test & ds.starts_with_open

mk_scatter(~ele_and_open & ~horizon_test, "both failures", "#7570b3")
mk_scatter(~ele_and_open & horizon_test, "just count${}^($ failure", "#d95f02")
mk_scatter(ele_and_open & ~horizon_test, "just horizon failure", "#e7298a")
mk_scatter(is_balanced, "balanced", "#1b9e77")

plt.xlabel("Logit difference from 2.0")
plt.ylabel("Logit difference from 2.1")
plt.legend()
plt.gcf().set_size_inches(5, 5)
plt.show()

# %%

fig, axs = plt.subplots(1, 3, sharey=True, figsize=(10, 3))
axs[0].set_title("Passes horizon test?")
pass_c = "#377eb8"
fail_c = "#e41a1c"
alpha = 0.3
size = 8

mk_scatter(horizon_test, color=pass_c, ax=axs[0])
mk_scatter(~horizon_test, color=fail_c, ax=axs[0])

axs[1].set_title("Passes count test?")
mk_scatter(~count_test, color=fail_c, ax=axs[1])
mk_scatter(count_test, color=pass_c, ax=axs[1])

axs[2].set_title("Passes count${}^($ test?")
mk_scatter(~ele_and_open, "fail", fail_c, ax=axs[2])
mk_scatter(ele_and_open, "pass", pass_c, ax=axs[2])

axs[2].legend()

axs[0].set_xlabel("Logit difference from 2.0")
axs[1].set_xlabel("Logit difference from 2.0")
axs[2].set_xlabel("Logit difference from 2.0")
axs[0].set_ylabel("Logit difference from 2.1")

plt.show()
# %%



import einops
import torch

import rust_circuit as rc

torch.manual_seed(0)


def test_conv():
    filter = rc.Array(torch.tensor([[[1], [1], [1]]], dtype=torch.float32))
    input = rc.Array.randn(2, 10, 1)
    print(input.shape, filter.shape)
    print("input", input)
    convy = rc.Conv(input, filter, stride=[1], padding=[(0, 1)])
    result = convy.evaluate()
    print("result", result)
    assert convy.shape == result.shape


def test_conv_2d():
    filter = torch.ones(11, 2, 2, 7)
    input = torch.randn(2, 3, 5, 7)
    raw_test_conv(input, filter, [1, 1], [(1, 1), (1, 1)])
    raw_test_conv(input, filter, [2, 1], [(0, 0), (0, 0)])
    raw_test_conv(torch.randn(10, 30, 40, 19), torch.randn(35, 5, 4, 19), [2, 1], [(0, 0), (0, 0)])
    raw_test_conv(torch.randn(10, 30, 40, 8), torch.randn(16, 5, 4, 8), [2, 1], [(0, 0), (0, 0)])


def raw_test_conv(input, filter, stride, padding):
    convy = rc.Conv(rc.Array(input), rc.Array(filter), stride, padding)
    rearranged_input = einops.rearrange(input, "a b c d -> a d b c")
    rearranged_filter = einops.rearrange(filter, "a b c d->a d b c")
    result = convy.evaluate()
    torch_result_rearranged = torch.nn.functional.conv2d(
        rearranged_input, rearranged_filter, None, stride, tuple([x[0] for x in padding])
    )
    torch_result = einops.rearrange(torch_result_rearranged, "a b c d -> a c d b")
    torch.testing.assert_close(result, torch_result)


if __name__ == "__main__":
    # testy1()
    test_conv()
    test_conv_2d()



from __future__ import annotations

import itertools
from typing import Callable, Iterable, List, Optional, Sequence, Tuple, TypeVar, cast

from rust_circuit import optional as op

from ._rust import (
    Add,
    Circuit,
    Cumulant,
    Einsum,
    Module,
    ModuleArgSpec,
    ModuleSpec,
    Scalar,
    Symbol,
    extract_rewrite_raw,
    kappa_term,
)


# source: https://stackoverflow.com/questions/19368375/set-partitions-in-python
def partition(collection: List[T]) -> Iterable[List[List[T]]]:
    if len(collection) == 0:
        yield []
        return

    if len(collection) == 1:
        yield [collection]
        return

    first = collection[0]
    for smaller in partition(collection[1:]):
        # insert `first` in each of the subpartition's subsets
        for n, subset in enumerate(smaller):
            yield smaller[:n] + [[first] + subset] + smaller[n + 1 :]
        # put `first` in its own subset
        yield [[first]] + smaller


def cum_name_to_eps_name(cum_name: str):
    """Replace the "k" at the beginning by an "eps" """
    return "eps" + cum_name[1:] if cum_name[0] == "k" else cum_name + "_eps"


def kappa_hat(cumulant: Cumulant):
    """this is an mathematical object we define for eps attrib/decomposition"""
    assert cumulant.num_children > 0
    if cumulant.num_children == 1:
        return Add.minus(cumulant.children[0], cumulant, name=f"{cum_name_to_eps_name(cumulant.name)}")
    return cumulant


def eps_term(args: List[List[Tuple[int, Circuit]]]):
    new_out, _ = kappa_term(args, on_sub_cumulant_fn=kappa_hat)
    # Note: autonaming + rename in kappa_hat will produce a nice name
    return new_out


def get_multiplier_default(p: List[List[Tuple[int, Circuit]]]) -> float:
    return (-1) ** sum(len(b) > 1 for b in p)


T = TypeVar("T")


def split_in_buckets(seq: Sequence[T], bucket_sizes: Sequence[int]) -> List[Sequence[T]]:
    res: List[Sequence[T]] = []
    runnning_count = 0
    for bsize in bucket_sizes:
        res.append(seq[runnning_count : runnning_count + bsize])
        runnning_count += bsize
    return res


def eps_attrib_module(
    node_ranks: Sequence[int],
    non_centered: bool = False,
    name: str = "epsilon",
    extra_filter: Callable[[List[List[Tuple[int, Circuit]]]], bool] = lambda _: True,
    get_multiplier: Callable[[List[List[Tuple[int, Circuit]]]], float] = get_multiplier_default,
    symbolic_size_start: int = 50,
) -> ModuleSpec:
    """returns epsilion(*circuits)
    non_centered means we don't subtract out the cumulant
    """

    shapes = split_in_buckets([None] * 1000, node_ranks)

    circuits: List[Circuit] = list[Circuit](
        [Symbol.new_with_random_uuid(tuple(shape), f"input_{i}") for i, shape in enumerate(shapes)]
    )
    input_specs = [(c, ModuleArgSpec(cast(Symbol, c))) for c in circuits]

    out: List[Einsum] = []

    if len(circuits) == 0:
        eps: Circuit = Scalar(1.0, name=name)
        return extract_rewrite_raw(eps, input_specs, prefix_to_strip=None, module_name=name).spec

    if non_centered and len(circuits) == 1:
        eps = circuits[0]
        return extract_rewrite_raw(eps, input_specs, prefix_to_strip=None, module_name=name).spec

    for p in partition(list(enumerate(circuits))):
        if (non_centered and len(p) == 1) or not extra_filter(p):
            continue

        new_out = eps_term(p)
        assert new_out.shape == tuple(itertools.chain.from_iterable(c.shape for c in circuits))
        out.append(Einsum.scalar_mul(new_out, get_multiplier(p)))
    eps = Add(*out, name=f"{name}_sum")

    return extract_rewrite_raw(eps, input_specs, prefix_to_strip=None, module_name=name).spec


def eps_attrib(
    cumulant: Cumulant,
    non_centered: bool = False,
    name: Optional[str] = None,
    extra_filter: Callable[[List[List[Tuple[int, Circuit]]]], bool] = lambda _: True,
    get_multiplier: Callable[[List[List[Tuple[int, Circuit]]]], float] = get_multiplier_default,
):
    """returns epsilion(*cumulant.circuits)
    non_centered means we don't subtract out the cumulant
    """

    name_ = op.unwrap_or(name, cum_name_to_eps_name(cumulant.name))
    cum_circuits = cumulant.children
    return Module.new_flat(
        eps_attrib_module(
            [c.rank for c in cum_circuits],
            non_centered=non_centered,
            extra_filter=extra_filter,
            get_multiplier=get_multiplier,
        ),
        *cum_circuits,
        name=name_,
    )



import importlib.util
import uuid
from pathlib import Path
from typing import Optional

import pytest
import torch

import rust_circuit as rc
from rust_circuit.rrfs import RRFS_DIR


def test_basic_funcs_correct():
    tester = rc.GeneralFunctionSpecTester(test_with_rand=True, start_shape_num=1)
    for name in [
        "sigmoid",
        "tanh",
        # "rsqrt",
        "gelu",
        "relu",
        "step",
        "reciprocal",
        "log_exp_p_1",
        "gaussian_pdf",
        "gaussian_cdf",
        "softmax",
        "log_softmax",
        # "q_from_qr",
        "min",
        "max",
    ]:
        tester.test_many_shapes(
            rc.GeneralFunction.new_by_name(rc.Symbol.new_with_random_uuid((3, 3, 3)), spec_name=name).spec
        )


def test_cast():
    arr = rc.Array.randn(1, 2, device_dtype=rc.TorchDeviceDtypeOp("cpu", "float16"))
    casty1 = rc.GeneralFunction.new_cast(
        arr, rc.TorchDeviceDtypeOp("cpu", torch.float16), rc.TorchDeviceDtypeOp("cpu", torch.float64)
    )
    assert casty1.device_dtype == rc.TorchDeviceDtypeOp("cpu", "float64")
    assert casty1.evaluate().dtype == torch.float64
    print(casty1, casty1.device_dtype, casty1.evaluate())

    with pytest.raises(rc.MiscInputCastIncompatibleDeviceDtypeError):
        casty2 = rc.GeneralFunction.new_cast(
            arr, rc.TorchDeviceDtypeOp("cpu", "int16"), rc.TorchDeviceDtypeOp("cpu", "float64")
        )
        print(casty2, casty2.device_dtype, casty2.evaluate())


@pytest.mark.cuda
def test_cast_cuda():
    arr = rc.Array.randn(1, 2, device_dtype=rc.TorchDeviceDtypeOp("cpu", "float16"))
    casty1 = rc.GeneralFunction.new_cast(
        arr, rc.TorchDeviceDtypeOp("cpu", "float16"), rc.TorchDeviceDtypeOp("cuda:0", "float64")
    )
    assert casty1.device_dtype == rc.TorchDeviceDtypeOp("cuda:0", "float64")
    assert casty1.evaluate().dtype == torch.float64


def test_custom_generalfunction_device_dtype():
    class MyCustomSigmoid(rc.GeneralFunctionSpecBase):
        def function(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore[override]
            return torch.sigmoid(x)

        def get_shape_info(self, *shapes: rc.Shape) -> rc.GeneralFunctionShapeInfo:
            return rc.get_shape_info_simple(shapes)

        # optionally you can implement `get_device_dtype_override`, which validates input device/dtypes and optionally returns the output device/dtype
        def get_device_dtype_override(self, *device_dtypes: rc.TorchDeviceDtypeOp) -> Optional[rc.TorchDeviceDtypeOp]:
            print("DOING OVERRIDE")
            assert device_dtypes[0].dtype in {
                "float16",
                "float32",
                "float64",
                "bfloat16",
            }  # softmax only applies to floats, not ints
            print("overrided")
            return None  # don't change device/dtype

    with pytest.raises(AssertionError):
        rc.GeneralFunction(rc.Array(torch.randint(10, 30, (2, 3))), spec=MyCustomSigmoid())
    rc.GeneralFunction(rc.Array(torch.randn(10, 30)), spec=MyCustomSigmoid())


def test_new_by_path():
    path = Path(RRFS_DIR) / "test" / "general_function.py"
    spec = importlib.util.spec_from_file_location("user_funcs", str(path))
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)
    spec = foo.MyCustomSigmoid()
    x = rc.Symbol.new_with_random_uuid((3, 3, 3))
    f_by_path = rc.GeneralFunction.new_by_path(x, path="/test/general_function.py:MyCustomSigmoid")
    f_by_spec = rc.GeneralFunction(x, spec=spec)
    assert f_by_path == f_by_spec
    assert f_by_path == rc.Parser()(f_by_path.repr())
    assert f_by_spec == rc.Parser()(f_by_spec.repr())



from uuid import uuid4

import attrs
import hypothesis
import hypothesis.extra.numpy as st_np
import torch

from interp.circuit.circuit import Circuit
from interp.circuit.circuit_compiler.util import RearrangeSpec
from interp.circuit.circuit_utils import cast_circuit, circuit_map, get_nodes
from interp.circuit.computational_node import Add, GeneralFunction, Index, UnaryRearrange
from interp.circuit.constant import ArrayConstant, One, Zero
from interp.circuit.get_update_node import FunctionIterativeNodeMatcher as F
from interp.circuit.get_update_node import NodeUpdater as NU
from interp.circuit.print_circuit import lambda_notation_circuit
from interp.circuit.testing.topdown_circuit import CircuitProbs as CP
from interp.circuit.testing.topdown_circuit import st_Circuit
from interp.circuit.var import AutoTag
from rust_circuit import Scalar as rScalar
from rust_circuit import Tag as rTag
from rust_circuit.interop_rust import py_to_rust


@hypothesis.settings(max_examples=200)
@hypothesis.given(
    circ=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_constant=False,
        must_be_explicitly_computable=False,
        growth_order="breadth_first",
        possibly_different_name=True,
        probs_default=CP.kw(
            all=1,
            Zero=0,
            One=0,
            GeneralFunction=0,
            Add=(1, dict(use_weights=False)),
            Index=(1, dict(allow_out_of_bounds_slice=False)),
        ),  # TODO: fix rearrange maybe
    ),
)
def test_rust_and_back(circ: Circuit):
    all_circs = get_nodes(circ)
    for c in all_circs:
        hypothesis.assume(not isinstance(c, (Zero, One, GeneralFunction)), "How did we get here???")
        if isinstance(c, Add):
            hypothesis.assume(all(x == 1.0 for x in c.items.values()), "How did we get here???")

    raw_test_rust_and_back(circ)


def raw_test_rust_and_back(circ: Circuit):
    circ = cast_circuit(circ, dtype=torch.float64)
    circ = circuit_map(
        lambda x: UnaryRearrange.from_spec(
            x.node,
            RearrangeSpec.from_rust(
                x.get_spec().to_rust().canonicalize().fill_empty_ints().conform_to_input_shape(x.node.shape)
            ),
            name=x.name,
        )
        if isinstance(x, UnaryRearrange)
        else x,
        circ,
    )
    circ_tagged = circuit_map(lambda x: AutoTag(x, x.uuid) if isinstance(x, ArrayConstant) else x, circ)
    rust_circ = py_to_rust(circ_tagged)
    back_py = rust_circ.to_py()
    back_py = NU(
        lambda x: attrs.evolve(x.node, uuid=x.uuid) if isinstance(x, AutoTag) else x,
        F(lambda x: isinstance(x, AutoTag) and isinstance(x.node, ArrayConstant)),
        assert_exists=False,
        assert_unique=False,
    )(back_py)
    back_py = circuit_map(
        lambda x: attrs.evolve(x, hash_tensor_idx_by_value=True) if isinstance(x, Index) else x, back_py
    )

    if back_py != circ:
        print(lambda_notation_circuit(back_py, tensors_as_randn=True))
        print(lambda_notation_circuit(circ, tensors_as_randn=True))
    assert back_py == circ


def test_auto_tag():
    assert rTag.new_with_random_uuid(rScalar(1)) != rTag.new_with_random_uuid(rScalar(1))
    common_uuid = uuid4()
    assert rTag(rScalar(1), common_uuid) == rTag(rScalar(1), common_uuid)



#%%
from rust_circuit import (
    FINISHED,
    Add,
    Circuit,
    IterateMatchResults,
    IterativeMatcher,
    Matcher,
    Scalar,
    add_flatten,
    new_traversal,
    restrict,
)

#%% [markdown]
# # Understanding Iterative Matchers
# ## 1. Matchers
# In brief, you can use them to find circuits in a parent circuit, regardless of their position in the parent.
# %%
# An example
parent = Add(Add(Scalar(3.14, name="pi"), Scalar(2.7, name="e")), Add(name="empty"))
parent.print()
# %%
# Here is how to find all circuits in the parent of type "Add"
Matcher(Add).get(parent)

# %%
# you can also do
parent.get(Add)
# which is equivalent to the above

# %%
# Here is how to find the circuit which name is "pi"
Matcher("pi").get_unique(parent)
# %%
# or
parent.get_unique("pi")
# %% [markdown]
# How does it work?
# A Matcher is a function: Circuit -> bool (with a few helper method to use it in other contexts), and ".get" while apply it to every circuit in the parent.
# %%
# You can match using many other attributes. See test_rust_get_update.py for more examples.
#%% [markdown]
# ## 2. IterativeMatchers (for finding circuits)
# In brief, you can use them to find circuits in a parent circuit, using their relation to the parent.
#%%
# Example 1: finding all Adds at depth 1 from the parent
restrict(IterativeMatcher(Add), start_depth=1, end_depth=2).get(parent)
# %%
# Example 2: finding the second child of the parent
restrict(IterativeMatcher(True), term_if_matches=True).children_matcher({1}).get(parent)
# Details:
# IterativeMatcher(True) -> match every node encountered
# .filter(term_if_matches=True) -> finish if match a node (will match only the parent)
# .children_matcher({1}) -> match the first child of every circuit matched by the callee
# Note: this is equivalent to `IterativeMatcher.term(match_next=True).children_matcher({1}).get(parent)`
#%%
# Example 3: Matching every Scalar, no matter their position (behaves like a Match)
IterativeMatcher(Scalar).get(parent)
# %% [markdown]
# How does it work?
# An IterativeMatcher is a function: Circuit -> IterateMatchResults (with a few helper method to use it in other contexts), where IterateMatchResults is an object holding 2 informations:
# - do we match the circuit? (named "found")
# - what do we do next? (named "updated")
# The IterativeMatcher is then called on the parent, and then it will propagate according to the answer to "what do we do next"
# %%
# what do we do next, option 1: finish
finish_immeditaly = IterativeMatcher.new_func(lambda circuit: IterateMatchResults(found=True, updated=FINISHED))
finish_immeditaly.get(parent)
# Details:
# - FINISHED is a constant provided by rust_circuit
# - to construct an IterativeMatcher by hand, we use IterativeMatcher.new_func, because we can't guess if a lambda returns a bool (which it would we passed a "Matcher"), or an IterateMatchResults.
# %%
# what do we do next, option 2: use the same matcher, for example here is the add matcher.
add_it_matcher = IterativeMatcher.new_func(
    lambda circuit: IterateMatchResults(found=isinstance(circuit, Add), updated=None)
)
add_it_matcher.get(parent)
# Detail: None means "use the same matcher"
# %%
# what do we do next, option 3: use a new IterativeMatcher for each child, for example, here is to return the second child of the first child

# this is equivalent to IterativeMatcher.term(match_next=True)
it_matcher_3 = IterativeMatcher.new_func(lambda _: IterateMatchResults(found=True, updated=FINISHED))

it_matcher_2 = IterativeMatcher.new_func(lambda _: IterateMatchResults(found=False, updated=[FINISHED, it_matcher_3]))
it_matcher_1 = IterativeMatcher.new_func(lambda _: IterateMatchResults(found=False, updated=[it_matcher_2, FINISHED]))
it_matcher_1.get(parent)
# %%
# You can also use the third option to store and modify a state in the IterativeMatcher
def countdown_factory(count: int) -> IterativeMatcher:
    def countdown_it_matcher(circuit: Circuit) -> IterateMatchResults:
        if count == 0:
            print("boom")
            return IterateMatchResults(FINISHED, found=True)
        if count % 2 == 0:
            print("tick")
        if count % 2 == 1:
            print("tack")
        updated_state = count - 1
        return IterateMatchResults([countdown_factory(updated_state) for _ in circuit.children], found=False)

    return IterativeMatcher.new_func(countdown_it_matcher)


chain = Add(name="root")
for i in range(100 + 1):
    chain = Add(chain, name=f"{i}")

countdown_factory(10).get_unique(chain).name
# %%
# ## 3. IterativeMatchers (for defining traversals)
# In brief, you can use them to define a subtree a parent circuit.
# Sometimes, it doesn't make sense to match any subset of the circuits of the parent.
# A traversal is what you will you give to a function need a subtree (that includes the root):
# it's just an IterativeMatcher, but instead of looking at what gets matched, the receiving function
# only looks at what get's traversed by the matcher until it finishes, and ignores the "found" term
# %%
# Example 1: flatten nested adds up to depth 2
one = Scalar(1.0, name="one")
parent = Add(
    Add(Add(one, name="bc"), Add(Add(one, name="aa"), name="c"), name="ab"),
    Add(Add(one, name="ae"), name="ad"),
    name="ac",
)
print("before")
parent.print()
# traversal is a constant equal to IterativeMatcher(True) (it can also be constructed via IterativeMatcher.noop_traversal())
my_traversal = new_traversal(end_depth=2 + 1)
print("after")
add_flatten(parent, my_traversal).print()
#%%
# The same thing happens if we match nothing instead of everything
my_traversal = restrict(IterativeMatcher(False), end_depth=2 + 1)
add_flatten(parent, my_traversal).print()
# %%
# Example 2: flatten nested adds that start with "a" connected to the root
def it_macher_fn(circuit: Circuit) -> IterateMatchResults:
    if not isinstance(circuit, Add):
        raise ValueError
    return IterateMatchResults(
        [
            (IterativeMatcher.new_func(it_macher_fn) if isinstance(c, Add) and c.name.startswith("a") else FINISHED)
            for c in circuit.children
        ]
    )  # Don't even explore nodes with don't start with "a"!


my_traversal = IterativeMatcher.new_func(it_macher_fn)
add_flatten(parent, my_traversal).print()
# %% [markdown]
# Traversals are also used for print, distribute, push_down_index, ...
# In general, you will not create the traversal by hand, and you will be able to use one using the different utilities at your disposal, either using filters, chains, ... or using a built-in traversal like traverse_until_depth.
# %% [markdown]
# ## 4. Updating nodes
# In addition to getting nodes we can also apply functions which update them
# %%

one = Scalar(1.0, name="one")
parent = Add(
    Add(Add(one, name="bc"), Add(Add(one, name="aa"), name="c"), name="ab"),
    Add(Add(one, name="ae"), name="ad"),
    name="ac",
)
print("before")
parent.print()
print()
print("after")
# update circuits with a 'c' anywhere in the name
updated = parent.update(Matcher.regex("c"), lambda x: x.rename(x.name + " has c"))
updated.print()

# %%

# you can also use
assert Matcher.regex("c").update(parent, lambda x: x.rename(x.name + " has c")) == updated

# %% [markdown]
# ## 5. Advanced subjects
#%% [markdown]
# You can chain iterators!
# When the head of the chain matches, the next element in the chain is started.
# When the tail of the chain matches, the matched circuit is considered matched by the chain.
# More info in the docstring for chain.
#%%
# Example 1: match all circuits which have a circuit "pi + e" as parent
parent = Add(Add(Scalar(3.14, name="pi"), Scalar(2.7, name="e")), Add(name="empty"))
Matcher("pi + e").chain(Matcher(True)).get(parent)
# Details:
# - using matcher.<some IterativeMatcher function> behaves as if the Matcher was case
#   to an IterativeMatcher which always returned updated=None
# - you can pass a Matcher constructor argument to the chain (this will construct the IterativeMatcher),
#   or a Matcher, or an IterativeMatcher.
#   This also works for most functions taking IterativeMatcher as argument. Read rust_circuit.pyi for more details.
#%%
# Example 2: return the second child of the first child
def n_th_child_iterative_matcher(n):
    return IterativeMatcher.term(True).children_matcher({n})


n_th_child_iterative_matcher(0).chain(n_th_child_iterative_matcher(1)).get(parent)
# %%



"""Check that all current notebooks and demos aren't broken with the current code.

You should be more willing than usual to remove tests from this file. If a notebook really isn't being used (for
research or documentation) any more, you should delete its test from here.

"""
import pytest

from rust_circuit.demos.notebook_testing import NotebookInTesting


def test_causal_scrubbing_paren_balancer():
    with NotebookInTesting():
        from rust_circuit.demos.paren_balancer.causal_scrubbing_experiments import __name__  # noqa: F401


def test_demo_rust_basic_scope_manager():
    with NotebookInTesting():
        from rust_circuit.demos.rust_circuit_demos.basic_scope_manager import __name__  # noqa: F401


def test_demo_rust_custom_general_function():
    with NotebookInTesting():
        from rust_circuit.demos.rust_circuit_demos.custom_general_function import __name__  # noqa: F401


def test_demo_rust_fancy_error():
    with NotebookInTesting():
        from rust_circuit.demos.rust_circuit_demos.fancy_error import __name__  # noqa: F401


@pytest.mark.skip(reason="broken demo, but the general NB testing should be on main")
def test_demo_rust_handcrafted_model_cumulants():
    with NotebookInTesting():
        from rust_circuit.demos.rust_circuit_demos.handcrafted_model_cumulants import __name__  # noqa: F401


def test_demo_rust_iterative_matchers():
    with NotebookInTesting():
        from rust_circuit.demos.rust_circuit_demos.iterative_matchers import __name__  # noqa: F401


def test_demo_rust_modules_and_symbols():
    with NotebookInTesting():
        from rust_circuit.demos.rust_circuit_demos.modules_and_symbols import __name__  # noqa: F401


def test_demo_rust_nest():
    with NotebookInTesting():
        from rust_circuit.demos.rust_circuit_demos.nest import __name__  # noqa: F401


def test_demo_rust_printing_parsing():
    with NotebookInTesting():
        from rust_circuit.demos.rust_circuit_demos.printing_parsing import __name__  # noqa: F401


def test_demo_rust_push_down_index():
    with NotebookInTesting():
        from rust_circuit.demos.rust_circuit_demos.push_down_index import __name__  # noqa: F401


def test_demo_rust_simp():
    with NotebookInTesting():
        from rust_circuit.demos.rust_circuit_demos.simp import __name__  # noqa: F401


def test_demo_rust_save_models():
    with NotebookInTesting():
        from rust_circuit.demos.rust_circuit_demos.save_models import __name__  # noqa: F401


def test_demo_push_down_module():
    with NotebookInTesting():
        from rust_circuit.demos.rust_circuit_demos.push_down_module import __name__  # noqa: F401


def test_demo_causal_scrubbing_simple():
    with NotebookInTesting():
        from rust_circuit.demos.causal_scrubbing.causal_scrubbing_simple import __name__  # noqa: F401


def test_demo_causal_scrubbing_pool():
    with NotebookInTesting():
        from rust_circuit.demos.causal_scrubbing.pool import __name__  # noqa: F401


def test_demo_transformer_config():
    with NotebookInTesting():
        from rust_circuit.demos.rust_circuit_demos.transformer_config import __name__  # noqa: F401



# %% [markdown]
# # Causal scrubbing
# %% [markdown]
# ## A simple example

import uuid

# We'll walk through how the code implements the causal scrubbing algorithm in the case of a simple model (it's the one we discuss in
# our [writeup](https://www.alignmentforum.org/posts/JvZhhzycHu2Yd57RN/causal-scrubbing-redwood-research#3_1_An_informal_description__What_activation_replacements_does_a_hypothesis_imply_are_valid_);
# though to run the code we have to make it a bit more specific).
# %% [markdown]
### G to explain and proposed correspondence
# %%
from typing import Dict, List, Tuple

import torch

import rust_circuit as rc
from rust_circuit.causal_scrubbing.dataset import color_dataset
from rust_circuit.causal_scrubbing.experiment import Experiment, ExperimentEvalSettings, ScrubbedExperiment
from rust_circuit.causal_scrubbing.hypothesis import (
    Correspondence,
    FuncSampler,
    InterpNode,
    corr_root_matcher,
    to_inputs,
)
from rust_circuit.causal_scrubbing.testing_utils import IntDataset
from rust_circuit.demos.notebook_testing import NotebookInTesting
from rust_circuit.py_utils import I

# %%
# Make a dataset of xs and labels
data_generator = torch.Generator()
data_generator.manual_seed(33)
data = torch.randint(high=10, size=(100, 3), generator=data_generator)
ds = IntDataset(
    (rc.Array(data, "xs"), rc.Array(torch.logical_or(data[:, 0] > 2, data[:, 1] > 4).to(torch.int64), "labels"))
)
print(repr(ds))
print(ds[0])

# Placeholder inputs; we'll feed in real inputs later
xs = rc.Array(torch.zeros((3,)), name="xs")
labels = rc.Array(torch.zeros(()), name="labels")

# Construct the computational graph (I just made this one up; in practice this might be some neural network).
# You might find it helpful to draw this graph on paper.
x0 = rc.Index(xs, I[0], name="x0")
x1 = rc.Index(xs, I[1], name="x1")
x2 = rc.Index(xs, I[2], name="x2")
A = rc.sigmoid(rc.Add(x0, rc.Einsum.scalar_mul(x1, 0), rc.Einsum.scalar_mul(x2, 0), rc.Scalar(-3)), name="A")
B = rc.sigmoid(rc.Add(x0, rc.Einsum.scalar_mul(x1, 0), rc.Einsum.scalar_mul(x2, 0), rc.Scalar(-3)), name="B")
C = rc.Add(x0, x1, x2)
D = rc.Add(rc.Einsum.elementwise_broadcasted(A, B), rc.Einsum.scalar_mul(C, 0), name="D")
loss = rc.Einsum.from_einsum_string(",  -> ", D, labels, name="loss")

loss.print()
# %%
# Defining our interpretation graph: we're going to hypothesize that e.g. the node "A" in our model
# only cares about whether the the data at index 0 is > 3 or not. Again, you might find it helpful
# to draw this out!
# First we define some helper functions:
def x0_f(d: IntDataset):
    return (d.xs.value)[:, 0]


def x1_f(d: IntDataset):
    return (d.xs.value)[:, 1]


def x0_gt_3(d: IntDataset):
    return x0_f(d) > 3


def x1_gt_3(d: IntDataset):
    return x1_f(d) > 3


def x0_or_x1_gt_3(d: IntDataset):
    return torch.logical_or(x0_gt_3(d), x1_gt_3(d))


def x0_or_x1_gt_3_agrees_with_label(d: IntDataset):
    return x0_or_x1_gt_3(d) == d.labels.value


def label_f(d: IntDataset):
    return d.labels.value


# Then we actually create a graph of InterpNodes
# FuncSampler of a function means that our sampled inputs must agree on this function with the reference;
# we'll talk more about CondSamplers lower down, for now you just need to know that you need to wrap
# your function in it.
out = InterpNode(cond_sampler=FuncSampler(x0_or_x1_gt_3_agrees_with_label), name="out")  # type: ignore
D_prime = out.make_descendant(FuncSampler(x0_or_x1_gt_3), name="D'")  # type: ignore
A_prime = D_prime.make_descendant(FuncSampler(x0_gt_3), name="A'")  # type: ignore
B_prime = D_prime.make_descendant(FuncSampler(x1_gt_3), name="B'")  # type: ignore
x0_prime = A_prime.make_descendant(FuncSampler(x0_f), name="x0'")  # type: ignore
x1_prime = B_prime.make_descendant(FuncSampler(x1_f), name="x1'")  # type: ignore
y_prime = out.make_descendant(FuncSampler(label_f), name="y'")  # type: ignore
# We can print out the graph we've defined!
out.print()
# %%
# And, finally, writing out the correspondence between the interpretation I (rooted at the InterpNode `out`)
# and the computational graph G (rooted at the Circuit `loss`).
# In particular, we associate every interp node with an IterativeMatcher which points at a part of the treeified circuit.
corr = Correspondence()
corr.add(out, corr_root_matcher)
corr.add(D_prime, rc.IterativeMatcher("D"))
corr.add(B_prime, rc.IterativeMatcher("B"))
corr.add(A_prime, rc.IterativeMatcher("A"))
corr.add(x1_prime, rc.IterativeMatcher("B").chain("x1"))
corr.add(x0_prime, rc.IterativeMatcher("A").chain("x0"))
corr.add(y_prime, rc.IterativeMatcher("labels"))
# %%

# %% [markdown]

# ### Doing causal scrubbing: recursively sampling data
#
# Terminology note: in circuits land, including here, we think of our circuit as being a dependency graph
# with arrows from nodes to nodes they depend on. This means that the output is an *ancestor* of the
# input. So, any time you see "parent", this means something closer to the output, and vice versa for "child".
# In the causal scrubbing algorithm, we traverse the interpretation graph from parents to children.
#
# The first thing we do in a causal scrubbing experiment is recursively (starting from the root, i.e. the output)
# sample two new data for each node of our interpretation graph. At each node n_I we have two inputs:
# 1. A source dataset that we can sample from
# 2. A parent datum that constrains our next sample
# And are sampling:
# 1. The new datum, which agrees with the parent datum on the function at n_I
# 2. An "other datum" that will be used for any inputs to this node that are not specified by the correspondence
#
# Normally this would happen behind the scenes in an Experiment, but here we'll walk through it step by step.

# After each recursive sampling step, we'll print out the interpretation. The nodes will be colored
# by the hash of their dataset, and they'll be annotated with the two data we've sampled for them.
# As you run the sampling, verify for yourself that the sampled datum at each node does in fact agree
# on the value of the function at that node with the datum of its parent.
# %%
generator = torch.Generator()  # rng generation
generator.manual_seed(22)

# the reference datum is the parent datum for the root; see section towards the end for why we do this
ref_datum = ds.sample(1, generator)
print(f"ref_datum: {str(ref_datum)}")
sampled_inputs = out.sample(generator, ds, ref_datum, recursive=False)
out.print(sampled_inputs=sampled_inputs)
# %%
def get_parents_and_children(i: InterpNode):
    return [(i, c) for c in i.children] + [pair for c in i.children for pair in get_parents_and_children(c)]


def step_sampler(parents_and_children: List[Tuple[InterpNode, InterpNode]]):
    for (p, c) in parents_and_children:
        c._sample_into(generator, ds, sampled_inputs[p], into=sampled_inputs, recursive=False)
        yield


step_sample = step_sampler(get_parents_and_children(out))
# %%
# we sample for D' a datum that agrees on if x0 > 3 OR x1 > 3
next(step_sample)
out.print()
# %%
# we sample for y' a datum that agrees on the label
next(step_sample)
out.print()
# %%
# we sample for A' a datum that agrees on if x0 > 3
next(step_sample)
out.print()
# %%
# we sample for B' a datum that agrees on if x1 > 3
next(step_sample)
out.print()
# %%
# we sample for x0' a datum that agrees on x0
next(step_sample)
out.print()
# %%
# we sample for x1' a datum that agrees on x1
next(step_sample)
out.print()
# %%
try:
    next(step_sample)
    assert False, "should be done sampling at this point!"
except StopIteration:
    pass

assert (  # To make sure this notebook doesn't get broken
    out.print(
        rc.PrintOptions(), sampled_inputs, color_by_data=False, repr=True
    )  # ignoring colors, which are sadly nondeterministic atm
    == """out GeneralFunction # cond_sampler=FuncSampler(d=IntDatum(xs=[9, 1, 0], label=1), f(d)=True), other_inputs_sampler=UncondSampler(d=IntDatum(xs=[0, 6, 8], label=1))
  D' GeneralFunction # cond_sampler=FuncSampler(d=IntDatum(xs=[0, 9, 8], label=1), f(d)=True), other_inputs_sampler=UncondSampler(d=IntDatum(xs=[3, 2, 9], label=1))
    A' GeneralFunction # cond_sampler=FuncSampler(d=IntDatum(xs=[0, 8, 5], label=1), f(d)=False), other_inputs_sampler=UncondSampler(d=IntDatum(xs=[0, 4, 6], label=0))
      x0' GeneralFunction # cond_sampler=FuncSampler(d=IntDatum(xs=[0, 5, 5], label=1), f(d)=0)
    B' GeneralFunction # cond_sampler=FuncSampler(d=IntDatum(xs=[0, 2, 8], label=0), f(d)=False), other_inputs_sampler=UncondSampler(d=IntDatum(xs=[0, 9, 4], label=1))
      x1' GeneralFunction # cond_sampler=FuncSampler(d=IntDatum(xs=[3, 9, 8], label=1), f(d)=9)
  y' GeneralFunction # cond_sampler=FuncSampler(d=IntDatum(xs=[6, 0, 6], label=1), f(d)=1)"""
)
# %% [markdown]

# ### Doing causal scrubbing: replacing inputs to the treeified model

# Recall that in causal scrubbing, we need to treeify our model; but we don't need to treeify it entirely,
# only as much as our hypothesis requires. That is, if our hypothesis says that some nodes need to be sampled
# together, they can stay together.

# In the context of an Experiment, we can call treeified() to perform this rewrite. This finds the paths
# that are picked out by this hypothesis, and wraps the inputs to each with a random tag to uniqueify them.
# This is an algebraic rewrite -- the resulting circuit is extensionally equal -- because we're not actually
# changing any of the inputs.
#
# Treeifying also allows us to run some checks that your correspondence is doing what it says on the tin.
# These checks assert various invariances that aren't otherwise enforced when constructing a Correspondence or
# Experiment, and catch various misleading correspondences that will run but give surprising results.
# We recommend running with treeify=True and check=True to validate your correspondence when you first create it.
#
# Before running the below, try drawing out what the treeified graph will look like on paper.
# %%
ex = Experiment(loss, ds, corr, random_seed=11)
treeified_circuit = ex.treeified()
corr.check(treeified_circuit, circuit_treeified=True)

treeified_circuit.print_html()
# %% [markdown]
# After that we can replace the inputs. Normally this would happen behind the scenes in an experiment, but we'll step through it here.
# %%
circuit = ex.wrap_in_var(treeified_circuit, ref_datum, rc.DiscreteVar.uniform_probs_and_group(1))

already_scrubbed: Dict[rc.IterativeMatcher, InterpNode] = {}
already_scrubbed_inputs: Dict[rc.IterativeMatcher, InterpNode] = {}


def colorer(c: rc.Circuit, super: rc.Circuit) -> str:
    color = "darkgrey"
    for m, i in already_scrubbed.items():
        m_endpoints = m.get(super)
        if c in m_endpoints:
            color = color_dataset(sampled_inputs.datasets[i], html=True)
        elif color == "darkgrey" and i.is_leaf() and c.are_any_found(m_endpoints):
            color = "lightgrey"
    for m, i in already_scrubbed_inputs.items():
        m_endpoints = m.get(super)
        if c in m_endpoints:
            color = color_dataset(sampled_inputs.datasets[i], html=True)
    return color  # type: ignore


def commenter(c: rc.Circuit, super: rc.Circuit) -> str:
    comment = ""
    for m, i in already_scrubbed.items():
        if c in m.get(super):
            comment = i.str_samplers(sampled_inputs)
    for m, i in already_scrubbed_inputs.items():
        if c in m.get(super):
            comment = f"ds set by '{i.name}': {sampled_inputs[i]}"
    return comment


po = rc.PrintHtmlOptions(
    colorer=lambda c: colorer(c, circuit),
    traversal=rc.IterativeMatcher.noop_traversal(),
    commenters=[lambda c: commenter(c, circuit)],
)


def step_scrubber(ex: Experiment, circuit: rc.Circuit):
    for interp_node, m in ex.nodes.in_dfs_order():
        print(interp_node)
        input_matcher = to_inputs(m, ex.dataset)
        already_scrubbed[m] = interp_node
        already_scrubbed_inputs[input_matcher] = interp_node
        # Internal method, called here just for demonstration
        circuit = ex._replace_one_input(circuit, interp_node, m, sampled_inputs)
        yield circuit


# %%
# Notice we're not using ScrubbedExperiment.print() here: the logic there assumes we have finished replacing all the
# inputs, so in order to color and annotate nodes as we step through we write the printing logic ourselves. As in
# ScrubbedExperiment.print(), nodes here are being colored to match their corresponding interp nodes. Dark grey nodes
# are outside the image of the correspondence (or we haven't gotten to replacing them yet), and light grey nodes are in
# matched paths but not directly mapped to. The annotation shows the data we sampled at that node, and the value of the
# interp node's function on that datum.
#
# These are exactly the data that we sampled earlier--we are now using those samples to replace the inputs to our
# circuit.
#
# You might find it useful to expand the nodes (click a > to turn it into a v) so you can see the entire treeified circuit, including
# the colors and annotations at the inputs. By default repeated nodes will be collapsed (including sometimes other copies of the input)

step_scrub = step_scrubber(ex, circuit)

circuit.print(po)
# %%
# We set _all_ inputs downstream of 'out' with out's 'other datum'
circuit = next(step_scrub)
circuit.print(po)
# %%
# We set all inputs downstream of 'D' with D's 'other datum'
circuit = next(step_scrub)
circuit.print(po)
# %%
# We set all inputs downstream of 'A' with A's 'other datum'
circuit = next(step_scrub)
circuit.print(po)
# %%
# We set all inputs downstream of 'x0'. As x0 is a leaf node we get to use it's main datset!
circuit = next(step_scrub)
circuit.print(po)
# %%
# Now into B, again with the 'other datum'
circuit = next(step_scrub)
circuit.print(po)
# %%
# B has one child -- x1. x1 is a leaf, so we use it's main data set to set inputs
circuit = next(step_scrub)
circuit.print(po)
# %%
# And finally set the inputs upstream of labels (also a leaf interp node)
circuit = next(step_scrub)
circuit.print(po)
# %%
try:
    next(step_scrub)
    assert False, "should be done scrubbing at this point!"
except StopIteration:
    pass

treeify_first = circuit
# %%
# Or, we can replace inputs without explicitly treeifying-- the treeification will happen automatically
# as inputs to different paths are replaced!
circuit = ex.base_circuit
group = rc.DiscreteVar.uniform_probs_and_group(1)
circuit = ex.wrap_in_var(circuit, ref_datum, group)
circuit = ex.replace_inputs(circuit, sampled_inputs)
circuit.print()

# %%
# Tags have random uuids, and treeified-first has some extra tags, but otherwise the two circuits are identical.
# We'll remove the extra tags, and replace all tags with the same uuid.
clean_tags = lambda c: c.update(
    lambda x: x.is_tag(), lambda x: rc.Tag(x.node, uuid.UUID("00000000-0000-0000-0000-000000000000"))
)
treeify_first_clean = clean_tags(treeify_first).update(lambda x: x.is_tag() and x.node.is_var(), lambda x: x.node)
assert treeify_first_clean == clean_tags(circuit)
# %% [markdown]

# ### Doing causal scrubbing: computing the behavior function on the scrubbed model
#
# Now we just run the scrubbed model forward!
# %%
device = "cpu" if NotebookInTesting.currently_in_notebook_test else "cuda:0"  # for CI
eval_settings = ExperimentEvalSettings(device_dtype=device)
scrubbed_out = ScrubbedExperiment(circuit, ref_datum, sampled_inputs, group, 11, ex._nodes).evaluate(
    eval_settings=eval_settings
)
print(scrubbed_out)
# %% [markdown]

# ### Doing causal scrubbing: expectation of the behavior function
# In practice, we run causal scrubbing on a batch and take the expectation.
# %%
out_batch = Experiment(loss, ds, corr, random_seed=11).scrub(20).evaluate(eval_settings)
print(out_batch.mean())
# %%
# Recall that our goal is to get as close as possible to the value of the behavior function on the original inputs:
loss_with_data = rc.Expander(("xs", lambda _: ds.xs), ("labels", lambda _: ds.labels))(loss)
print(loss_with_data.evaluate().mean())
# %%
# %% [markdown]

# ## Conveniences
#
# The causal scrubbing code provides a number of conveniences to make expressing your hypotheses easier, rather than forcing you
# to manually specify a graph isomorphism between your interpretation and the model.
# %% [markdown]

# ### IterativeMatcher

# Suppose you had the diamond graph (where A is the input and D is the output)
# ```
#   A
#  / \
# B   C
#  \ /
#   D
# ```
# and you wanted to hypothesize that A does some computation and later D uses that information to do something. You might want to
# draw the interpretation `A -> D`; but there is no such edge in the model, so you can't write down a homomorphism c from I to G!
#
# Or, suppose you were analyzing a neural network. You had analyzed the attention heads and thought 0.1 and 2.1 were important while
# no heads in layer 1 were, but you had not yet looked at the MLPs. How would you express the hypothesis that paths from 0.1 to 2.1,
# potentially passing through MLPs but not through layer 1 heads, were important?
#
# In both of these cases, you could rewrite your interpretation with more nodes -- in the first example, adding nodes to your
# interpretation corresponding to B and C. Or, you could do a circuit rewrite to group together some of the nodes in your model.
# Both of these are pretty annoying to do, so instead we allow you to correspond an interpretation node to an IterativeMatcher picking
# out some paths in the model. That is, you can say things like:
# ```
# correspondence.add(a_I, rc.IterativeMatcher("D").chain("A"))
# ```

# %% [markdown]

# ### Implicit extension to inputs

# If you have not explicitly specified an interpretation for some of the paths between c(I) and the inputs, this is filled
# in for you. After all, we have to have inputs for every path through the model! In other words, for every matcher in your
# interp graph, all the inputs upstream of nodes that matcher matches will be replaced. They will be replaced by the interp
# node's `ds` if they are leaves, and its `other_inputs_ds` if they are not (see the next section); some inputs will then be
# replaced again by child interp nodes if they exist.
#
# To do this we have to know what the "inputs" are. The names of the inputs are defined in the dataset you supply to the
# experiment, and are expected to be present in the circuit you pass in.

# %% [markdown]

# ### More on: nodes not in any matched paths
#
# We mentioned briefly earlier that each interpretation node n_I samples a other_inputs_ds. If c(n_I) has children that are not in
# the paths defined by c(children(n_I)), they will be run on this dataset. For instance, in the example we looked at, the input to
# x1 and x2 going into A was drawn from A_primes other_inputs_ds. This allows you to very easily say that "the other
# children don't matter" -- by default, other_inputs_ds is sampled randomly from the source dataset.
#
# But you can also use this in other ways: for example, you can set n_I's cond_sampler to an UncondSampler, and n_I's other_inputs_sampler
# to an ExactSampler, which represents "c(children(n_I)) don't matter, and the other children do matter and can't be resampled at all".
# For example, you can use this to ablate a single path in the model.
#
# ```
# n_I = InterpNode(cond_sampler=UncondSampler(), other_inputs_sampler=ExactSampler(), name="n")
# ```
#
# Sometimes we might call the set `children(c(n_I)) \ c(children(n_I))` the "unimportant" children as a shorthand, even though they
# are not necessarily unimportant. We also sometimes refer to them as "not in the image of c".

# %% [markdown]

# ### Ref ds for extending the graph
#
# When you .evaluate a scrubbed experiment, you only get the output on the circuit that you specified. If the scrubbed
# circuit computes your entire behavior function, you are done. But, if you want to feed the result of your scrubbed circuit into
# a continued behavior function, you should use the reference ds for any computations there, which you can get by reading the
# ScrubbedExperiment.ref_ds field. For example, if the circuit computes logits and you want to compute a loss, you should use the
# reference dataset to get the labels.

# %% [markdown]

# ## Checks
#
# The code performs a number of checks to make sure your correspondence induces well-defined scrubs. For examples, take a look
# at test_hypothesis.py and test_experiment.py. But we'll briefly list out the checks here as well:
#
# - out maps to out
# - in maps to in
# - corr must map every node in the interp graph to something
# - must respect model tree structure: an interp node I is a child of I' iff each path in c(I) is an extension of a path in c(I')
# - each matcher must match something
# - no two matchers can match the same path
# - bonus: after treeifying, no two matchers can match the same node. this should be equivalent to the above.
#
# Most of these won't run if you disable checks on your experiment; but some are mandatory (like the fact that each matcher must
# match something).
# %% [markdown]

# ## Generalization
#
# Our code is also more general than the algorithm we describe in our writeup in a couple of useful ways.
#
# 1. Each interp node, rather than computing a function and sampling data that agree with the parent datum on that function,
# is associated with a sampler. This is convenient for lots of reasons, like generating data on the fly, but is also fundamentally
# more general: you do not have to sample uniformly from a subset of the data!
# 2. The interpretation is not a computational graph. Again, some things are simply more convenient to express as conditional on the
# dataset rather than the outputs of a computational graph--for example, sampling some things together, enforcing agreement on a feature
# the samples for different nodes, phi rewrites. But, conditions are more powerful than agreement on a function; we won't elaborate on
# that here.

# %% [markdown]

# ## Code base components
#
# The causal scrubbing code relies on the following components:
# 1. An `Experiment`, which takes a `Circuit`, a `Dataset`, and a `Correspondence`.
# 2. A `Circuit`, the model we want to interpret
# 3. A `Correspondence` object, which associates `InterpNodes` with `IterativeMatchers`
# 4. `InterpNodes` are associated in a tree structure, where children are created by calling
# `InterpNode().make_decendant()`. Each `InterpNode` has a name (str) and two `CondSamplers`, one
# that is used for it's 'main datum' and one used for the 'other datum'
# 5. `CondSamplers` are objects that implement `sample(ref_datum, dataset) -> new_datum`.
# For instance a `FuncSampler` will sample new_datum from dataset that agrees on a specified function.
# In reality, they do this in a vectorized way from `(ref_dataset, dataset) -> new_dataset`
# but it can be helpful to think one datapoint at a time.
# 6. `IterativeMatcher` which points at a single subcircuit of the model, potentially on a restricted set of paths.
# 7. `Dataset` objects are pretty simple, mostly wrappers around several arrays with some
# convenience functions (e.g. printing). We require that they frozen for hashing.

# %%



# %%
import torch

import rust_circuit as rc

# %%

# We can print circuits, like so:

circuit = rc.Einsum.from_einsum_string(
    "ab,bc->aa",
    rc.Scalar(1.0, (2, 3), "scalar1"),
    rc.Rearrange.from_string(rc.Scalar(2.0, (4, 3), "scalar2"), "a b -> b a"),
    name="einy",
)
circuit.print()

# repr is same as printing, just to string
assert (
    circuit.repr()
    == """
'einy' Einsum ab,bc->aa
  'scalar1' [2,3] Scalar 1
  'scalar2 rearrange' Rearrange a b -> b a
    'scalar2' [4,3] Scalar 2
""".strip()
)

# python's print(circuit) by default only prints depth 2 (root and children, no grandchildren), and wraps it in <> so you it works well with python datastructures
print((circuit, 1))
# %%

# Or use options

rc.PrintOptions(bijection=True).print(circuit)

# if a subcircuit appears multiple times,
# subsequent appearances will be replaced with just the name or serial number
assert (
    rc.Add(rc.Scalar(1.0, (2,)), rc.Scalar(1.0, (2,))).repr()
    == """
0 Add
  1 [2] Scalar 1
  1 Scalar""".strip()
)
assert (
    rc.Add(rc.Scalar(1.0, (2,), "scaley"), rc.Scalar(1.0, (2,), "scaley"), name="addey").repr()
    == """
\'addey' Add
  'scaley' [2] Scalar 1
  'scaley'""".strip()
)

# %%

# if bijection=True (which is the default!), you can parse printed circuits back into circuits again! This is lossless, except for custom GeneralFunctions

P = rc.Parser()
parsed = P(rc.PrintOptions(bijection=True).repr(circuit))
assert parsed == circuit

# You can also write circuits yourself, and store them in files called .circ
P(
    """
'mycircuit' Einsum a,ab->b
  'sym1' [4] Symbol 
  'sym2' [4,5] Symbol 
"""
)

# (indentation width is 2 spaces, not configurable right now)

# you can use parse_circuits to parse a file containing multiple circuits
# you can use this with names/serial numbers to factor out parts of your circuit
# parsing also supports comments with #


_, factored_circuit = P.parse_circuits(
    """
# the submodule, declared first
'submod1' [2] Einsum a,a->a
  'inp' [2] Scalar 1
  'inp'
'main_circuit' Einsum a,aa->
  'submod1' # the reference to the submodule
  'traced' [2,2] Scalar 2
"""
)
factored_circuit.print()

# %%

# rust_circuit hashes all tensors (contents and type info) on intake
# Tensors are stored by hash, locally (by default in ~/tensors_by_hash_cache automatically) and in rrfs/tensor_db
# in a fixed file trie.

# when you print a circuit with bijection=True, if any of its tensors that aren't saved locally will be saved
# later you can sync your local tensors to rrfs with rc.sync_all_unsynced_tensors
# or you can use rc.PrintOptions(sync_tensors=True) to print and push tensors to rrfs immediately (slow).
# There's also sync_specific_tensors(...) which takes a list of hash prefix strings

demo_array = rc.Array(torch.zeros(2, 3))
demo_array.print()
circuit_string_to_share = rc.PrintOptions(bijection=True, sync_tensors=True).repr(
    P(
        """
0 Add
  1 [2,3] Array 2f97b0bb06d4adba46cd8756
  1
  """
    )
)
print(circuit_string_to_share)

# Here are the formats for all the node info types
# shapes are only provided when needed. Only Scalar, Array, Symbol, Scatter need shapes
P.parse_circuits(
    """
'_' [] Scalar 1 # any number
'_1' [2] Symbol 93aa0db1-2435-4fa6-8675-711543cab705 # no uuid means null uuid
'_2' [2,3] Array 2f97b0bb06d4adba46cd8756 # prefix of hash of tensor
'einsum' Einsum a,ab,->b
  '_1'
  '_2'
  '_'
'einsum fancy string' Einsum fancy: ay, ay bee,->bee
  '_1'
  '_2'
  '_'
'add' Add
  '_1'
  '_'
'rearrange' Rearrange a b -> b a ()
  '_2'
'index' Index [0:1,1]
  '_2'
'cumulant' Cumulant
  '_'
  '_1'
"""
)

# %%
# Options rundown:

rc.PrintOptions(arrows=True).print(circuit)
# Parsing treats all characters used in arrows as whitespace, so you can parse with arrows=True (but not control nesting with arrows, that's indentation)

# colorers! you can supply a function, or use a built in one
rc.PrintOptions(colorer=rc.PrintOptions.hash_colorer()).print(circuit)
rc.PrintOptions(colorer=rc.PrintOptions.type_colorer()).print(circuit)

# commenters is a list of functions that generate comment strings, put after the node in the same line
rc.PrintOptions(commenters=[lambda x: str(rc.count_nodes(x))]).print(circuit)

# this prints child argnames (such as einsum axes) as comments on children
rc.PrintOptions(comment_arg_names=True).print(
    P(
        """
0 DiscreteVar
  1 Einsum ab,b->ab
    's1' [2,3] Symbol
    's2' [3] Symbol
  's3' [2] Symbol
"""
    )
)


# %%

# You can terminate printing early (avoid printing children) with a traversal. (this is incompatible with bijeciton=True)

deepcirc = P(
    """
0 Rearrange ->
  1 Rearrange ->
    2 Rearrange ->
      'sym' [] Symbol
"""
)
rc.PrintOptions(bijection=False, traversal=rc.new_traversal(start_depth=0, end_depth=1)).print(deepcirc)
rc.PrintOptions(bijection=False, traversal=rc.new_traversal(start_depth=0, end_depth=2)).print(deepcirc)

# Diff print

orig_circuit = P(
    """
'a0' Add
  'a1' Add
    'sym1' [2] Symbol
    'sym2' [2] Symbol
  'a2' Add
    'sym1'
    'sym2'
    'sym2'
"""
)
rewritten_circuit = orig_circuit.update("sym2", lambda _: P("'sym3' [2] Symbol"))
rewritten_circuit.print()
print(rc.diff_circuits(rewritten_circuit, orig_circuit, rc.PrintOptions()))

rewritten_circuit = orig_circuit.update("sym2", lambda _: P("'sc0' [2] Scalar 1"))
rewritten_circuit.print()
print(rc.diff_circuits(rewritten_circuit, orig_circuit, rc.PrintOptions()))

# %%
